'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import math
from copy import deepcopy

from pm4py.objects.dfg.utils import dfg_utils
from pm4py.util import constants, nx_utils

DEFAULT_NOISE_THRESH_DF = 0.16


def generate_nx_graph_from_dfg(dfg, start_activities, end_activities, activities_count):
    """
    Generate a NetworkX graph for reachability-checking purposes out of the DFG

    Parameters
    --------------
    dfg
        DFG
    start_activities
        Start activities
    end_activities
        End activities
    activities_count
        Activities of the DFG along with their count

    Returns
    --------------
    G
        NetworkX digraph
    start_node
        Identifier of the start node (connected to all the start activities)
    end_node
        Identifier of the end node (connected to all the end activities)
    """
    start_node = '4d872045-8664-4e21-bd55-5da5edb096fe' # made static to avoid undeterminism
    end_node = 'b8136db7-b162-4763-bd68-4d5ccbcdff87' # made static to avoid undeterminism
    G = nx_utils.DiGraph()
    G.add_node(start_node)
    G.add_node(end_node)
    for act in activities_count:
        G.add_node(act)
    for edge in dfg:
        G.add_edge(edge[0], edge[1])
    for act in start_activities:
        G.add_edge(start_node, act)
    for act in end_activities:
        G.add_edge(act, end_node)
    return G, start_node, end_node


def filter_dfg_on_activities_percentage(dfg0, start_activities0, end_activities0, activities_count0, percentage):
    """
    Filters a DFG (complete, and so connected) on the specified percentage of activities
    (but ensuring that every node is still reachable from the start and to the end)

    Parameters
    ----------------
    dfg0
        (Complete, and so connected) DFG
    start_activities0
        Start activities
    end_activities0
        End activities
    activities_count0
        Activities of the DFG along with their count
    percentage
        Percentage of activities

    Returns
    ----------------
    dfg
        (Filtered) DFG
    start_activities
        (Filtered) start activities
    end_activities
        (Filtered) end activities
    activities_count
        (Filtered) activities of the DFG along with their count
    """
    # since the dictionaries/sets are modified, a deepcopy is the best option to ensure data integrity
    dfg = deepcopy(dfg0)
    start_activities = deepcopy(start_activities0)
    end_activities = deepcopy(end_activities0)
    activities_count = deepcopy(activities_count0)

    if len(activities_count) > 1 and len(dfg) > 1:
        activities_count_sorted_list = sorted([(x, y) for x, y in activities_count.items()], key=lambda x: (x[1], x[0]),
                                              reverse=True)
        # retrieve the minimum list of activities to keep in the graph, according to the percentage
        min_set_activities_to_keep = set(
            x[0] for x in activities_count_sorted_list[:math.ceil((len(activities_count) - 1) * percentage) + 1])
        # retrieve the activities that can be possibly discarded, according to the percentage
        activities_to_possibly_discard = list(
            x[0] for x in activities_count_sorted_list[math.ceil((len(activities_count) - 1) * percentage) + 1:])
        activities_to_possibly_discard.reverse()
        # build a graph structure that helps in deciding whether the activities can be discarded safely
        graph, start_node, end_node = generate_nx_graph_from_dfg(dfg, start_activities, end_activities,
                                                                 activities_count)
        for act in activities_to_possibly_discard:
            new_graph = nx_utils.DiGraph(graph)
            # try to remove the node
            new_graph.remove_node(act)
            # check whether all the activities to keep can be reached from the start and can reach the end
            reachable_from_start = set(nx_utils.descendants(new_graph, start_node))
            reachable_to_end = set(nx_utils.ancestors(new_graph, end_node))
            if min_set_activities_to_keep.issubset(reachable_from_start) and min_set_activities_to_keep.issubset(
                    reachable_to_end):
                # if that is the case, try to elaborate the new DFG (without the activity)
                new_dfg = {x: y for x, y in dfg.items() if x[0] != act and x[1] != act}
                # if that is still not empty ...
                if new_dfg:
                    # ... then the activity can be safely removed
                    dfg = new_dfg
                    del activities_count[act]
                    if act in start_activities:
                        del start_activities[act]
                    if act in end_activities:
                        del end_activities[act]
                    graph = new_graph

        # at the end of the previous step, some nodes may be remaining that are not reachable from the start
        # or cannot reach the end. obviously the previous steps ensured that at least the activities in min_set_activities_to_keep
        # are connected
        reachable_from_start = set(nx_utils.descendants(graph, start_node))
        reachable_to_end = set(nx_utils.ancestors(graph, end_node))
        reachable_start_end = reachable_from_start.intersection(reachable_to_end)
        activities_set = set(activities_count.keys())
        non_reachable_activities = activities_set.difference(reachable_start_end)

        # remove these non reachable activities
        for act in non_reachable_activities:
            dfg = {x: y for x, y in dfg.items() if x[0] != act and x[1] != act}
            del activities_count[act]
            if act in start_activities:
                del start_activities[act]
            if act in end_activities:
                del end_activities[act]

    return dfg, start_activities, end_activities, activities_count


def __filter_specified_paths(dfg, start_activities, end_activities, activities_count, graph, start_node, end_node,
                             discardable_edges, activities_not_to_discard):
    for edge in discardable_edges:
        if len(dfg) > 1:
            new_graph = nx_utils.DiGraph(graph)
            # try to remove the edge
            new_graph.remove_edge(edge[0], edge[1])

            # check whether all the activities to keep can be reached from the start and can reach the end
            reachable_from_start = set(nx_utils.descendants(new_graph, start_node))
            reachable_to_end = set(nx_utils.ancestors(new_graph, end_node))

            if activities_not_to_discard.issubset(reachable_from_start) and activities_not_to_discard.issubset(
                    reachable_to_end):
                # remove the edge
                graph = new_graph
                if edge in dfg:
                    # if the edge connects two activities simply remove that
                    del dfg[edge]
                elif edge[0] == start_node:
                    del start_activities[edge[1]]
                elif edge[1] == end_node:
                    del end_activities[edge[0]]

    # at the end of the previous step, some nodes may be remaining that are not reachable from the start
    # or cannot reach the end. obviously the previous steps ensured that at least the activities in min_set_activities_to_keep
    # are connected
    reachable_from_start = set(nx_utils.descendants(graph, start_node))
    reachable_to_end = set(nx_utils.ancestors(graph, end_node))
    reachable_start_end = reachable_from_start.intersection(reachable_to_end)
    activities_set = set(activities_count.keys())
    non_reachable_activities = activities_set.difference(reachable_start_end)

    # remove these non reachable activities
    for act in non_reachable_activities:
        dfg = {x: y for x, y in dfg.items() if x[0] != act and x[1] != act}
        del activities_count[act]
        if act in start_activities:
            del start_activities[act]
        if act in end_activities:
            del end_activities[act]

    # make sure that the DFG contains only edges between these activities
    dfg = {x: y for x, y in dfg.items() if x[0] in activities_count and x[1] in activities_count}

    return dfg, start_activities, end_activities, activities_count


def filter_dfg_on_paths_percentage(dfg0, start_activities0, end_activities0, activities_count0, percentage,
                                   keep_all_activities=False):
    """
    Filters a DFG (complete, and so connected) on the specified percentage of paths
    (but ensuring that every node is still reachable from the start and to the end)

    Parameters
    ----------------
    dfg0
        (Complete, and so connected) DFG
    start_activities0
        Start activities
    end_activities0
        End activities
    activities_count0
        Activities of the DFG along with their count
    percentage
        Percentage of paths
    keep_all_activities
        Decides if all the activities (also the ones connected by the low occurrences edges) should be kept,
        or only the ones appearing in the edges with more occurrences (default).

    Returns
    ----------------
    dfg
        (Filtered) DFG
    start_activities
        (Filtered) start activities
    end_activities
        (Filtered) end activities
    activities_count
        (Filtered) activities of the DFG along with their count
    """

    # since the dictionaries/sets are modified, a deepcopy is the best option to ensure data integrity
    dfg = deepcopy(dfg0)
    start_activities = deepcopy(start_activities0)
    end_activities = deepcopy(end_activities0)
    activities_count = deepcopy(activities_count0)

    if len(activities_count) > 1 and len(dfg) > 1:
        # build a graph structure that helps in deciding whether the paths can be discarded safely
        graph, start_node, end_node = generate_nx_graph_from_dfg(dfg, start_activities, end_activities,
                                                                 activities_count)
        all_edges = [(x, y) for x, y in dfg.items()] + [((start_node, x), start_activities[x]) for x in
                                                        start_activities] + [((x, end_node), end_activities[x]) for x in
                                                                             end_activities]
        all_edges = sorted(all_edges, key=lambda x: (x[1], x[0]), reverse=True)
        # calculate a set of edges that could be discarded and not
        non_discardable_edges = list(
            x[0] for x in all_edges[:math.ceil((len(all_edges) - 1) * percentage) + 1])
        discardable_edges = list(x[0] for x in all_edges[math.ceil((len(all_edges) - 1) * percentage) + 1:])
        discardable_edges.reverse()

        # according to the parameter's value, keep the activities that appears in the edges that should not be
        # discarded (default), OR keep all the activities, trying to remove edges but ensure connectiveness of
        # everything
        if keep_all_activities:
            activities_not_to_discard = set(x[0] for x in dfg).union(set(x[1] for x in dfg)).union(
                set(start_activities)).union(set(end_activities)).union(set(activities_count))
        else:
            activities_not_to_discard = set(x[0] for x in non_discardable_edges if not x[0] == start_node).union(
                set(x[1] for x in non_discardable_edges if not x[1] == end_node))

        dfg, start_activities, end_activities, activities_count = __filter_specified_paths(dfg, start_activities,
                                                                                           end_activities,
                                                                                           activities_count, graph,
                                                                                           start_node, end_node,
                                                                                           discardable_edges,
                                                                                           activities_not_to_discard)

    return dfg, start_activities, end_activities, activities_count


def filter_dfg_keep_connected(dfg0, start_activities0, end_activities0, activities_count0, threshold,
                              keep_all_activities=False):
    """
    Filters a DFG (complete, and so connected) on the specified dependency threshold (Heuristics Miner dependency)
    (but ensuring that every node is still reachable from the start and to the end)

    Parameters
    ----------------
    dfg0
        (Complete, and so connected) DFG
    start_activities0
        Start activities
    end_activities0
        End activities
    activities_count0
        Activities of the DFG along with their count
    threshold
        Dependency threshold as in the Heuristics Miner
    keep_all_activities
        Decides if all the activities should be kept,
        or only the ones appearing in the edges with higher threshold (default).

    Returns
    ----------------
    dfg
        (Filtered) DFG
    start_activities
        (Filtered) start activities
    end_activities
        (Filtered) end activities
    activities_count
        (Filtered) activities of the DFG along with their count
    """

    # since the dictionaries/sets are modified, a deepcopy is the best option to ensure data integrity
    dfg = deepcopy(dfg0)
    start_activities = deepcopy(start_activities0)
    end_activities = deepcopy(end_activities0)
    activities_count = deepcopy(activities_count0)

    if len(activities_count) > 1 and len(dfg) > 1:
        # build a graph structure that helps in deciding whether the paths can be discarded safely
        graph, start_node, end_node = generate_nx_graph_from_dfg(dfg, start_activities, end_activities,
                                                                 activities_count)

        dependency = {}
        for el in dfg:
            elinv = (el[1], el[0])
            if elinv in dfg:
                dep = (dfg[el] - dfg[elinv]) / (dfg[el] + dfg[elinv] + 1)
            else:
                dep = dfg[el] / (dfg[el] + 1)
            dependency[el] = dep

        all_edges = [(x, y) for x, y in dependency.items()] + [((start_node, x), 1.0) for x in
                                                        start_activities] + [((x, end_node), 1.0) for x in
                                                                             end_activities]
        all_edges = sorted(all_edges, key=lambda x: (x[1], x[0]), reverse=True)
        # calculate a set of edges that could be discarded and not
        non_discardable_edges = list(x[0] for x in all_edges if x[1] >= threshold)
        discardable_edges = list(x[0] for x in all_edges if x[1] < threshold)
        discardable_edges.reverse()

        # according to the parameter's value, keep the activities that appears in the edges that should not be
        # discarded (default), OR keep all the activities, trying to remove edges but ensure connectiveness of
        # everything
        if keep_all_activities:
            activities_not_to_discard = set(x[0] for x in dfg).union(set(x[1] for x in dfg)).union(
                set(start_activities)).union(set(end_activities)).union(set(activities_count))
        else:
            activities_not_to_discard = set(x[0] for x in non_discardable_edges if not x[0] == start_node).union(
                set(x[1] for x in non_discardable_edges if not x[1] == end_node))

        dfg, start_activities, end_activities, activities_count = __filter_specified_paths(dfg, start_activities,
                                                                                           end_activities,
                                                                                           activities_count, graph,
                                                                                           start_node, end_node,
                                                                                           discardable_edges,
                                                                                           activities_not_to_discard)

    return dfg, start_activities, end_activities, activities_count


def filter_dfg_to_activity(dfg0, start_activities0, end_activities0, activities_count0, target_activity,
                           parameters=None):
    """
    Filters the DFG, making "target_activity" the only possible end activity of the graph

    Parameters
    ---------------
    dfg0
        Directly-follows graph
    start_activities0
        Start activities
    end_activities0
        End activities
    activities_count0
        Activities count
    target_activity
        Target activity (only possible end activity after the filtering)
    parameters
        Parameters

    Returns
    ---------------
    dfg
        Filtered DFG
    start_activities
        Filtered start activities
    end_activities
        Filtered end activities
    activities_count
        Filtered activities count
    """
    if parameters is None:
        parameters = {}

    # since the dictionaries/sets are modified, a deepcopy is the best option to ensure data integrity
    dfg = deepcopy(dfg0)
    start_activities = deepcopy(start_activities0)
    activities_count = deepcopy(activities_count0)

    dfg = {x: y for x, y in dfg.items() if x[0] != target_activity}
    end_activities = {target_activity: activities_count[target_activity]}

    changed = True
    while changed:
        changed = False
        predecessors = dfg_utils.get_predecessors(dfg, activities_count)
        successors = dfg_utils.get_successors(dfg, activities_count)

        successors_from_sa = set()
        for act in start_activities:
            successors_from_sa = successors_from_sa.union(successors[act])
            successors_from_sa.add(act)

        reachable_nodes = successors_from_sa.intersection(predecessors[target_activity]).union({target_activity})
        if reachable_nodes != set(activities_count.keys()):
            changed = True
            activities_count = {x: y for x, y in activities_count.items() if x in reachable_nodes}
            start_activities = {x: y for x, y in start_activities.items() if x in reachable_nodes}
            dfg = {x: y for x, y in dfg.items() if x[0] in reachable_nodes and x[1] in reachable_nodes}

    return dfg, start_activities, end_activities, activities_count


def filter_dfg_from_activity(dfg0, start_activities0, end_activities0, activities_count0, source_activity,
                             parameters=None):
    """
    Filters the DFG, making "source_activity" the only possible source activity of the graph

    Parameters
    ---------------
    dfg0
        Directly-follows graph
    start_activities0
        Start activities
    end_activities0
        End activities
    activities_count0
        Activities count
    source_activity
        Source activity (only possible start activity after the filtering)
    parameters
        Parameters

    Returns
    ---------------
    dfg
        Filtered DFG
    start_activities
        Filtered start activities
    end_activities
        Filtered end activities
    activities_count
        Filtered activities count
    """
    if parameters is None:
        parameters = {}

    # since the dictionaries/sets are modified, a deepcopy is the best option to ensure data integrity
    dfg = deepcopy(dfg0)
    end_activities = deepcopy(end_activities0)
    activities_count = deepcopy(activities_count0)

    dfg = {x: y for x, y in dfg.items() if x[1] != source_activity}
    start_activities = {source_activity: activities_count[source_activity]}

    changed = True
    while changed:
        changed = False
        predecessors = dfg_utils.get_predecessors(dfg, activities_count)
        successors = dfg_utils.get_successors(dfg, activities_count)

        predecessors_from_ea = set()
        for ea in end_activities:
            predecessors_from_ea = predecessors_from_ea.union(predecessors[ea])
            predecessors_from_ea.add(ea)

        reachable_nodes = predecessors_from_ea.intersection(successors[source_activity]).union({source_activity})
        if reachable_nodes != set(activities_count.keys()):
            changed = True
            activities_count = {x: y for x, y in activities_count.items() if x in reachable_nodes}
            end_activities = {x: y for x, y in end_activities.items() if x in reachable_nodes}
            dfg = {x: y for x, y in dfg.items() if x[0] in reachable_nodes and x[1] in reachable_nodes}

    return dfg, start_activities, end_activities, activities_count


def filter_dfg_contain_activity(dfg0, start_activities0, end_activities0, activities_count0, activity, parameters=None):
    """
    Filters the DFG keeping only nodes that can reach / are reachable from activity

    Parameters
    ---------------
    dfg0
        Directly-follows graph
    start_activities0
        Start activities
    end_activities0
        End activities
    activities_count0
        Activities count
    activity
        Activity that should be reachable / should reach all the nodes of the filtered graph
    parameters
        Parameters

    Returns
    ---------------
    dfg
        Filtered DFG
    start_activities
        Filtered start activities
    end_activities
        Filtered end activities
    activities_count
        Filtered activities count
    """
    if parameters is None:
        parameters = {}

    # since the dictionaries/sets are modified, a deepcopy is the best option to ensure data integrity
    dfg = deepcopy(dfg0)
    start_activities = deepcopy(start_activities0)
    end_activities = deepcopy(end_activities0)
    activities_count = deepcopy(activities_count0)

    changed = True
    while changed:
        changed = False
        predecessors = dfg_utils.get_predecessors(dfg, activities_count)
        successors = dfg_utils.get_successors(dfg, activities_count)

        predecessors_act = predecessors[activity].union({activity})
        successors_act = successors[activity].union({activity})

        start_activities1 = {x: y for x, y in start_activities.items() if x in predecessors_act}
        end_activities1 = {x: y for x, y in end_activities.items() if x in successors_act}

        if start_activities != start_activities1 or end_activities != end_activities1:
            changed = True

        start_activities = start_activities1
        end_activities = end_activities1

        reachable_nodes = predecessors_act.union(successors_act)
        if reachable_nodes != set(activities_count.keys()):
            changed = True
            activities_count = {x: y for x, y in activities_count.items() if x in reachable_nodes}
            dfg = {x: y for x, y in dfg.items() if x[0] in reachable_nodes and x[1] in reachable_nodes}

    return dfg, start_activities, end_activities, activities_count


def clean_dfg_based_on_noise_thresh(dfg, activities, noise_threshold, parameters=None):
    """
    Clean Directly-Follows graph based on noise threshold

    Parameters
    ----------
    dfg
        Directly-Follows graph
    activities
        Activities in the DFG graph
    noise_threshold
        Noise threshold

    Returns
    ----------
    newDfg
        Cleaned dfg based on noise threshold
    """
    if parameters is None:
        parameters = {}

    most_common_paths = parameters[
        constants.PARAM_MOST_COMMON_PATHS] if constants.PARAM_MOST_COMMON_PATHS in parameters else None
    if most_common_paths is None:
        most_common_paths = []

    new_dfg = None
    activ_max_count = {}
    for act in activities:
        activ_max_count[act] = dfg_utils.get_max_activity_count(dfg, act)

    for el in dfg:
        if type(el[0]) is str:
            if new_dfg is None:
                new_dfg = {}
            act1 = el[0]
            act2 = el[1]
            val = dfg[el]
        else:
            if new_dfg is None:
                new_dfg = []
            act1 = el[0][0]
            act2 = el[0][1]
            val = el[1]

        if not el in most_common_paths and val < min(activ_max_count[act1] * noise_threshold,
                                                     activ_max_count[act2] * noise_threshold):
            pass
        else:
            if type(el[0]) is str:
                new_dfg[el] = dfg[el]
                pass
            else:
                new_dfg.append(el)
                pass

    if new_dfg is None:
        return dfg

    return new_dfg



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel.exporter.sqlite.variants import pandas_exporter, ocel20
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any


class Variants(Enum):
    PANDAS_EXPORTER = pandas_exporter
    OCEL20 = ocel20


def apply(ocel: OCEL, target_path: str, variant=Variants.PANDAS_EXPORTER, parameters: Optional[Dict[Any, Any]] = None):
    """
    Exports an OCEL to a SQLite database

    Parameters
    -------------
    ocel
        Object-centric event log
    target_path
        Path to the SQLite database
    variant
        Variant to use. Possible values:
        - Variants.PANDAS_EXPORTER => Pandas exporter
    parameters
        Variant-specific parameters
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(ocel, target_path, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from copy import deepcopy
from typing import Optional, Dict, Any, List, Set

from pm4py.objects.log.obj import Trace
from pm4py.objects.process_tree.utils import bottomup
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.objects.process_tree.utils.generic import fold
from pm4py.util import constants, xes_constants, exec_utils
from enum import Enum


class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def apply(tree: ProcessTree, trace: Trace, parameters: Optional[Dict[Any, Any]] = None, **kwargs) -> ProcessTree:
    """
    Reduce a process tree replacing the skippable elements that have empty intersection with the
    trace.

    Parameters
    -----------------
    tree
        Process tree
    trace
        Trace of an event log
    parameters
        Parameters of the algorithm, possible values: Parameters.ACTIVITY_KEY

    Returns
    ------------------
    tree
        Reduced process tree
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    activities = set(x[activity_key] for x in trace)

    tree = deepcopy(tree)
    from pm4py.algo.discovery.footprints.tree.variants import bottomup as footprints

    bottomup_nodes = bottomup.get_bottomup_nodes(tree)
    fps = footprints.get_all_footprints(tree)
    fps = {id(x): y for x, y in fps.items()}

    return reduce(bottomup_nodes, fps, activities)


def reduce(bottomup_nodes: List[ProcessTree], fps: Dict[str, Any], activities: Set[str]) -> ProcessTree:
    """
    Reduce a process tree replacing the skippable elements that have empty intersection with the
    trace.

    Parameters
    -----------------
    bottomup_nodes
        List of nodes of the process tree (that are process trees by themselves) in a bottomup order
    fps
        Footprints of the process tree
    activities
        Set of activities in the trace

    Returns
    ------------------
    tree
        Reduced process tree
    """
    i = 0
    while i < len(bottomup_nodes) - 1:
        node = bottomup_nodes[i]
        parent = node.parent

        is_skippable = fps[id(node)][Outputs.SKIPPABLE.value]
        node_activities = fps[id(node)][Outputs.ACTIVITIES.value]

        if is_skippable and not node_activities.intersection(activities):
            pt = ProcessTree()
            pt.parent = parent
            parent.children[parent.children.index(node)] = pt
        i = i + 1

    return fold(bottomup_nodes[-1])



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Set, Tuple
from copy import copy


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> Set[Tuple[str, str]]:
    """
    Calculates the object descendants graph.
    Two objects o1 and o2, both related to an event e, are connected if:
    - e is the last event of the lifecycle of o1
    - e is the first event of the lifecycle of o2
    
    Parameters
    -----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm
    
    Returns
    -----------------
    object_inheritance_graph
        Object inheritance graph (directed)
    """
    if parameters is None:
        parameters = {}

    graph = set()

    ordered_events = ocel.events[ocel.event_id_column].to_numpy().tolist()
    ev_rel_obj = ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column].agg(list).to_dict()
    last_event_per_obj = {}
    set_objects = set()

    ordered_events_revert = copy(ordered_events)
    ordered_events_revert.reverse()

    for ev in ordered_events_revert:
        rel_obj = ev_rel_obj[ev]
        rel_obj_seen = {x for x in rel_obj if x in set_objects}
        rel_obj_unseen = {x for x in rel_obj if x not in rel_obj_seen}

        for obj in rel_obj_unseen:
            last_event_per_obj[obj] = ev
            set_objects.add(obj)

    set_objects = set()

    for ev in ordered_events:
        rel_obj = ev_rel_obj[ev]
        rel_obj_seen = {x for x in rel_obj if x in set_objects}
        rel_obj_unseen = {x for x in rel_obj if x not in rel_obj_seen}
        rel_obj_last = {x for x in rel_obj if last_event_per_obj[x] == ev}

        for o2 in rel_obj_unseen:
            for o1 in rel_obj_last:
                if o1 != o2:
                    graph.add((o1, o2))
            set_objects.add(o2)

    graph_it = list(graph)
    for el in graph_it:
        if (el[1], el[0]) in graph:
            graph.remove((el[0], el[1]))
            graph.remove((el[1], el[0]))

    return graph



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.algo.discovery.footprints.tree.variants import bottomup as bottomup_discovery
from pm4py.objects.process_tree.utils import bottomup as bottomup_util
from enum import Enum

class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


FP_DEV_COLOR = "red"
FP_START_END_DEV_COLOR = "blue"
FP_ALWAYS_EXECUTED_DEV_COLOR = "orange"

FOOTPRINTS_KEY = "footprints"
START_ACTIVITIES = "start_activities"
END_ACTIVITIES = "end_activities"
ALWAYS_EXECUTED = "always_executed"


def apply(tree, conf_results, parameters=None):
    """
    Projects conformance results on top of the process tree

    Parameters
    --------------
    tree
        Process tree
    conf_results
        Conformance results (footprints on the entire log vs entire model)
    parameters
        Parameters of the algorithm

    Returns
    --------------
    color_map
        Color map to be provided to the visualization
    """
    if parameters is None:
        parameters = {}

    start_activities = {}
    end_activities = {}

    if isinstance(conf_results, list):
        raise Exception("the visualization can only be applied with total footprints (not trace-by-trace)!")
    elif isinstance(conf_results, dict):
        footprints = conf_results
        start_activities = conf_results[START_ACTIVITIES]
        end_activities = conf_results[END_ACTIVITIES]
    else:
        footprints = conf_results

    bottomup_nodes = bottomup_util.get_bottomup_nodes(tree, parameters=parameters)
    labels_dictio = {x.label: x for x in bottomup_nodes if x.operator is None and x.label is not None}
    all_fp_dictio = bottomup_discovery.get_all_footprints(tree, parameters=parameters)
    conf_colors = {}

    for res in start_activities:
        if res in labels_dictio:
            conf_colors[labels_dictio[res]] = FP_START_END_DEV_COLOR

    for res in end_activities:
        if res in labels_dictio:
            conf_colors[labels_dictio[res]] = FP_START_END_DEV_COLOR

    for res in footprints:
        if res[0] in labels_dictio and res[1] in labels_dictio:
            conf_colors[labels_dictio[res[0]]] = FP_DEV_COLOR
            conf_colors[labels_dictio[res[1]]] = FP_DEV_COLOR
            for n in bottomup_nodes:
                if res[0] in all_fp_dictio[n][Outputs.ACTIVITIES.value] and res[1] in all_fp_dictio[n][
                    Outputs.ACTIVITIES.value]:
                    conf_colors[n] = FP_DEV_COLOR
                    break

    return conf_colors



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import datetime
from copy import copy
from enum import Enum
from random import choice
from typing import Optional, Dict, Any, Union

from pm4py.objects import petri_net
from pm4py.objects.log import obj as log_instance
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import constants
from pm4py.util import exec_utils
from pm4py.util import xes_constants
from pm4py.util.dt_parsing.variants import strpfromiso


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    RETURN_VISITED_ELEMENTS = "return_visited_elements"
    NO_TRACES = "noTraces"
    MAX_TRACE_LENGTH = "maxTraceLength"
    PETRI_SEMANTICS = "petri_semantics"
    ADD_ONLY_IF_FM_IS_REACHED = "add_only_if_fm_is_reached"
    FM_LEQ_ACCEPTED = "fm_leq_accepted"
    INITIAL_TIMESTAMP = "initial_timestamp"
    INITIAL_CASE_ID = "initial_case_id"


def apply_playout(net, initial_marking, no_traces=100, max_trace_length=100,
                  initial_timestamp=10000000, initial_case_id=0,
                  case_id_key=xes_constants.DEFAULT_TRACEID_KEY,
                  activity_key=xes_constants.DEFAULT_NAME_KEY, timestamp_key=xes_constants.DEFAULT_TIMESTAMP_KEY,
                  final_marking=None, return_visited_elements=False, semantics=petri_net.semantics.ClassicSemantics(),
                  add_only_if_fm_is_reached=False, fm_leq_accepted=False):
    """
    Do the playout of a Petrinet generating a log

    Parameters
    ----------
    net
        Petri net to play-out
    initial_marking
        Initial marking of the Petri net
    no_traces
        Number of traces to generate
    max_trace_length
        Maximum number of events per trace (do break)
    initial_timestamp
        Increased timestamp from 1970 for the first event
    initial_case_id
        Case id of the first event
    case_id_key
        Trace attribute that is the case ID
    activity_key
        Event attribute that corresponds to the activity
    timestamp_key
        Event attribute that corresponds to the timestamp
    final_marking
        If provided, the final marking of the Petri net
    semantics
        Semantics of the Petri net to be used (default: petri_net.semantics.ClassicSemantics())
    add_only_if_fm_is_reached
        Adds the case only if the final marking is reached
    fm_leq_accepted
        Accepts traces ending in a marking that is a superset of the final marking
    """
    # assigns to each event an increased timestamp from 1970
    curr_timestamp = initial_timestamp
    all_visited_elements = []

    i = 0
    while True:
        if len(all_visited_elements) >= no_traces:
            break

        if i >= no_traces:
            if not add_only_if_fm_is_reached:
                break

            if len(all_visited_elements) == 0:
                # likely, the final marking is not reachable, therefore terminate here the playout
                break

        visited_elements = []
        visible_transitions_visited = []

        marking = copy(initial_marking)
        while len(visible_transitions_visited) < max_trace_length:
            visited_elements.append(marking)

            if not semantics.enabled_transitions(net, marking):  # supports nets with possible deadlocks
                break
            all_enabled_trans = semantics.enabled_transitions(net, marking)
            if final_marking is not None and final_marking <= marking and (final_marking == marking or fm_leq_accepted):
                trans = choice(list(all_enabled_trans.union({None})))
            else:
                trans = choice(list(all_enabled_trans))
            if trans is None:
                break

            visited_elements.append(trans)
            if trans.label is not None:
                visible_transitions_visited.append(trans)

            marking = semantics.execute(trans, net, marking)

        if not add_only_if_fm_is_reached:
            all_visited_elements.append(tuple(visited_elements))
        elif final_marking == marking:
            all_visited_elements.append(tuple(visited_elements))
        elif final_marking <= marking and fm_leq_accepted:
            all_visited_elements.append(tuple(visited_elements))

        i = i + 1

    if return_visited_elements:
        return all_visited_elements

    log = log_instance.EventLog()

    for index, visited_elements in enumerate(all_visited_elements):
        trace = log_instance.Trace()
        trace.attributes[case_id_key] = str(index+initial_case_id)
        for element in visited_elements:
            if type(element) is PetriNet.Transition and element.label is not None:
                event = log_instance.Event()
                event[activity_key] = element.label
                event[timestamp_key] = strpfromiso.fix_naivety(datetime.datetime.fromtimestamp(curr_timestamp))
                trace.append(event)
                # increases by 1 second
                curr_timestamp += 1
        log.append(trace)

    return log


def apply(net: PetriNet, initial_marking: Marking, final_marking: Marking = None,
          parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> EventLog:
    """
    Do the playout of a Petrinet generating a log

    Parameters
    -----------
    net
        Petri net to play-out
    initial_marking
        Initial marking of the Petri net
    final_marking
        If provided, the final marking of the Petri net
    parameters
        Parameters of the algorithm:
            Parameters.NO_TRACES -> Number of traces of the log to generate
            Parameters.MAX_TRACE_LENGTH -> Maximum trace length
            Parameters.INITIAL_TIMESTAMP -> The first event is set with INITIAL_TIMESTAMP increased from 1970
            Parameters.INITIAL_CASE_ID -> Numeric case id for the first trace
            Parameters.PETRI_SEMANTICS -> Petri net semantics to be used (default: petri_nets.semantics.ClassicSemantics())
            Parameters.ADD_ONLY_IF_FM_IS_REACHED -> adds the case only if the final marking is reached
            Parameters.FM_LEQ_ACCEPTED -> Accepts traces ending in a marking that is a superset of the final marking
    """
    if parameters is None:
        parameters = {}
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, xes_constants.DEFAULT_TRACEID_KEY)
    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    no_traces = exec_utils.get_param_value(Parameters.NO_TRACES, parameters, 1000)
    max_trace_length = exec_utils.get_param_value(Parameters.MAX_TRACE_LENGTH, parameters, 1000)
    initial_timestamp = exec_utils.get_param_value(Parameters.INITIAL_TIMESTAMP, parameters, 10000000)
    initial_case_id = exec_utils.get_param_value(Parameters.INITIAL_CASE_ID, parameters, 0)
    return_visited_elements = exec_utils.get_param_value(Parameters.RETURN_VISITED_ELEMENTS, parameters, False)
    semantics = exec_utils.get_param_value(Parameters.PETRI_SEMANTICS, parameters, petri_net.semantics.ClassicSemantics())
    add_only_if_fm_is_reached = exec_utils.get_param_value(Parameters.ADD_ONLY_IF_FM_IS_REACHED, parameters, False)
    fm_leq_accepted = exec_utils.get_param_value(Parameters.FM_LEQ_ACCEPTED, parameters, False)

    return apply_playout(net, initial_marking, max_trace_length=max_trace_length,
                         initial_timestamp=initial_timestamp,
                         initial_case_id=initial_case_id,
                         no_traces=no_traces,
                         case_id_key=case_id_key, activity_key=activity_key, timestamp_key=timestamp_key,
                         final_marking=final_marking, return_visited_elements=return_visited_elements,
                         semantics=semantics, add_only_if_fm_is_reached=add_only_if_fm_is_reached,
                         fm_leq_accepted=fm_leq_accepted)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.util import exec_utils
from enum import Enum
import tempfile
from graphviz import Digraph
from typing import Optional, Dict, Any
from pm4py.objects.bpmn.obj import BPMN
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from pm4py.util import constants
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    RANKDIR = "rankdir"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"


def add_bpmn_node(graph, n, font_size):
    n_id = str(id(n))

    if isinstance(n, BPMN.Task):
        graph.node(n_id, shape="box", label=n.get_name(), fontsize=font_size)
    elif isinstance(n, BPMN.StartEvent):
        graph.node(n_id, label="", shape="circle", style="filled", fillcolor="green", fontsize=font_size)
    elif isinstance(n, BPMN.EndEvent):
        graph.node(n_id, label="", shape="circle", style="filled", fillcolor="orange", fontsize=font_size)
    elif isinstance(n, BPMN.ParallelGateway):
        graph.node(n_id, label="+", shape="diamond", fontsize=font_size)
    elif isinstance(n, BPMN.ExclusiveGateway):
        graph.node(n_id, label="X", shape="diamond", fontsize=font_size)
    elif isinstance(n, BPMN.InclusiveGateway):
        graph.node(n_id, label="O", shape="diamond", fontsize=font_size)
    else:
        graph.node(n_id, label="", shape="circle", fontsize=font_size)


def apply(bpmn_graph: BPMN, parameters: Optional[Dict[Any, Any]] = None) -> graphviz.Digraph:
    """
    Visualize a BPMN graph

    Parameters
    -------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters of the visualization, including:
         - Parameters.FORMAT: the format of the visualization
         - Parameters.RANKDIR: the direction of the representation (default: LR)

    Returns
    ------------
    gviz
        Graphviz representation
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.bpmn.obj import BPMN
    from pm4py.objects.bpmn.util.sorting import get_sorted_nodes_edges

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ)
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 12)
    font_size = str(font_size)
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR)

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    filename.close()

    viz = Digraph("", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})
    viz.graph_attr['rankdir'] = rankdir

    nodes, edges = get_sorted_nodes_edges(bpmn_graph)
    process_ids = []
    for n in nodes:
        if n.process not in process_ids:
            process_ids.append(n.process)
    process_ids_members = {n.process: list() for n in nodes}
    for n in nodes:
        process_ids_members[n.process].append(n)

    for n in nodes:
        add_bpmn_node(viz, n, font_size)

    """
    viz.node('@@anchor', style='invis')

    for subp in process_ids:
        with viz.subgraph(name="cluster"+subp) as c:
            for n in process_ids_members[subp]:
                c.attr(label=subp)
                add_bpmn_node(c, n, font_size)
                c.attr(rank='same')

        viz.edge('@@anchor', str(id(process_ids_members[subp][0])), style='invis')
    """
    
    for e in edges:
        n_id_1 = str(id(e[0]))
        n_id_2 = str(id(e[1]))

        viz.edge(n_id_1, n_id_2)

    viz.attr(overlap='false')

    viz.format = image_format.replace("html", "plain-ext")

    return viz


def save(gviz: graphviz.Digraph, output_file_path: str, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)
    return ""


def view(gviz: graphviz.Digraph, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz: graphviz.Digraph, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """

    return gview.matplotlib_view(gviz, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from enum import Enum
from pm4py.util import exec_utils, constants, xes_constants
from pm4py.objects.ocel import constants as ocel_constants
from collections import Counter
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.statistics.ocel import objects_ot_count
from pm4py.objects.ocel.util import filtering_utils
from copy import copy


class Parameters(Enum):
    EVENT_ID = ocel_constants.PARAM_EVENT_ID
    OBJECT_ID = ocel_constants.PARAM_OBJECT_ID
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE


def apply(ocel: OCEL, min_num_obj_type: Dict[str, int], parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Filters the events of the object-centric logs which are related to at least
    the specified amount of objects per type.

    E.g. apply(ocel, {"order": 1, "element": 2})

    Would keep the following events:

      ocel:eid ocel:timestamp ocel:activity ocel:type:element ocel:type:order
    0       e1     1980-01-01  Create Order  [i4, i1, i3, i2]            [o1]
    1      e11     1981-01-01  Create Order          [i6, i5]            [o2]
    2      e14     1981-01-04  Create Order          [i8, i7]            [o3]

    Parameters
    ------------------
    ocel
        Object-centric event log
    min_num_obj_type
        Minimum number of objects per type
    parameters
        Parameters of the filter, including:
        - Parameters.EVENT_ID => the event identifier
        - Parameters.OBJECT_ID => the object identifier
        - Parameters.OBJECT_TYPE => the object type

    Returns
    -----------------
    filtered_event_log
        Filtered object-centric event log
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(Parameters.EVENT_ID, parameters, ocel.event_id_column)
    object_id = exec_utils.get_param_value(Parameters.OBJECT_ID, parameters, ocel.object_id_column)
    object_type = exec_utils.get_param_value(Parameters.OBJECT_TYPE, parameters, ocel.object_type_column)

    num_obj = objects_ot_count.get_objects_ot_count(ocel, parameters=parameters)

    filt_evs = set()

    for evid, evobjs in num_obj.items():
        is_ok = True
        for k, v in min_num_obj_type.items():
            if not k in evobjs:
                is_ok = False
                break
            elif evobjs[k] < v:
                is_ok = False
                break
        if is_ok:
            filt_evs.add(evid)

    ocel = copy(ocel)
    ocel.events = ocel.events[ocel.events[event_id].isin(filt_evs)]

    return filtering_utils.propagate_event_filtering(ocel, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import uuid

from lxml import etree

from pm4py.objects.petri_net.obj import Marking
from pm4py.objects.petri_net.obj import PetriNet, ResetNet, InhibitorNet
from pm4py.objects.petri_net import properties as petri_properties
from pm4py.util import constants, exec_utils
from enum import Enum


class Parameters(Enum):
    ENCODING = "encoding"


def export_petri_tree(petrinet, marking, final_marking=None, export_prom5=False, parameters=None):
    """
    Export a Petrinet to a XML tree

    Parameters
    ----------
    petrinet: :class:`pm4py.entities.petri.petrinet.PetriNet`
        Petri net
    marking: :class:`pm4py.entities.petri.petrinet.Marking`
        Marking
    final_marking: :class:`pm4py.entities.petri.petrinet.Marking`
        Final marking (optional)
    export_prom5
        Enables exporting PNML files in a format that is ProM5-friendly
    parameters
        Other parameters of the algorithm

    Returns
    ----------
    tree
        XML tree
    """
    if parameters is None:
        parameters = {}

    if final_marking is None:
        final_marking = Marking()

    root = etree.Element("pnml")
    net = etree.SubElement(root, "net")
    net.set("id", petrinet.name)
    netname = etree.SubElement(net, "name")
    netnametext = etree.SubElement(netname, "text")
    netnametext.text = petrinet.name
    net.set("type", "http://www.pnml.org/version-2009/grammar/pnmlcoremodel")
    if export_prom5 is True:
        page = net
    else:
        page = etree.SubElement(net, "page")
        page.set("id", "n0")
    places_map = {}
    for place in petrinet.places:
        places_map[place] = place.name
        pl = etree.SubElement(page, "place")
        pl.set("id", place.name)
        pl_name = etree.SubElement(pl, "name")
        pl_name_text = etree.SubElement(pl_name, "text")
        pl_name_text.text = place.properties[
            constants.PLACE_NAME_TAG] if constants.PLACE_NAME_TAG in place.properties else place.name
        if place in marking:
            pl_initial_marking = etree.SubElement(pl, "initialMarking")
            pl_initial_marking_text = etree.SubElement(pl_initial_marking, "text")
            pl_initial_marking_text.text = str(marking[place])
        if constants.LAYOUT_INFORMATION_PETRI in place.properties:
            graphics = etree.SubElement(pl, "graphics")
            position = etree.SubElement(graphics, "position")
            position.set("x", str(place.properties[constants.LAYOUT_INFORMATION_PETRI][0][0]))
            position.set("y", str(place.properties[constants.LAYOUT_INFORMATION_PETRI][0][1]))
            dimension = etree.SubElement(graphics, "dimension")
            dimension.set("x", str(place.properties[constants.LAYOUT_INFORMATION_PETRI][1][0]))
            dimension.set("y", str(place.properties[constants.LAYOUT_INFORMATION_PETRI][1][1]))
    transitions_map = {}
    for transition in petrinet.transitions:
        transitions_map[transition] = transition.name
        trans = etree.SubElement(page, "transition")
        trans.set("id", transition.name)
        trans_name = etree.SubElement(trans, "name")
        trans_text = etree.SubElement(trans_name, "text")
        if constants.LAYOUT_INFORMATION_PETRI in transition.properties:
            graphics = etree.SubElement(trans, "graphics")
            position = etree.SubElement(graphics, "position")
            position.set("x", str(transition.properties[constants.LAYOUT_INFORMATION_PETRI][0][0]))
            position.set("y", str(transition.properties[constants.LAYOUT_INFORMATION_PETRI][0][1]))
            dimension = etree.SubElement(graphics, "dimension")
            dimension.set("x", str(transition.properties[constants.LAYOUT_INFORMATION_PETRI][1][0]))
            dimension.set("y", str(transition.properties[constants.LAYOUT_INFORMATION_PETRI][1][1]))
        if constants.STOCHASTIC_DISTRIBUTION in transition.properties:
            random_variable = transition.properties[constants.STOCHASTIC_DISTRIBUTION]
            stochastic_information = etree.SubElement(trans, "toolspecific")
            stochastic_information.set("tool", "StochasticPetriNet")
            stochastic_information.set("version", "0.2")
            distribution_type = etree.SubElement(stochastic_information, "property")
            distribution_type.set("key", "distributionType")
            distribution_type.text = random_variable.get_distribution_type()
            if not random_variable.get_distribution_type() == "IMMEDIATE":
                distribution_parameters = etree.SubElement(stochastic_information, "property")
                distribution_parameters.set("key", "distributionParameters")
                distribution_parameters.text = random_variable.get_distribution_parameters()
            distribution_priority = etree.SubElement(stochastic_information, "property")
            distribution_priority.set("key", "priority")
            distribution_priority.text = str(random_variable.get_priority())
            distribution_invisible = etree.SubElement(stochastic_information, "property")
            distribution_invisible.set("key", "invisible")
            distribution_invisible.text = str(True if transition.label is None else False).lower()
            distribution_weight = etree.SubElement(stochastic_information, "property")
            distribution_weight.set("key", "weight")
            distribution_weight.text = str(random_variable.get_weight())
        if transition.label is not None:
            trans_text.text = transition.label
        else:
            trans_text.text = transition.name
            tool_specific = etree.SubElement(trans, "toolspecific")
            tool_specific.set("tool", "ProM")
            tool_specific.set("version", "6.4")
            tool_specific.set("activity", "$invisible$")
            tool_specific.set("localNodeID", str(uuid.uuid4()))
        if export_prom5 is True:
            if transition.label is not None:
                prom5_specific = etree.SubElement(trans, "toolspecific")
                prom5_specific.set("tool", "ProM")
                prom5_specific.set("version", "5.2")
                log_event_prom5 = etree.SubElement(prom5_specific, "logevent")
                event_name = transition.label.split("+")[0]
                event_transition = transition.label.split("+")[1] if len(
                    transition.label.split("+")) > 1 else "complete"
                log_event_prom5_name = etree.SubElement(log_event_prom5, "name")
                log_event_prom5_name.text = event_name
                log_event_prom5_type = etree.SubElement(log_event_prom5, "type")
                log_event_prom5_type.text = event_transition
        # specific for data Petri nets
        if petri_properties.TRANS_GUARD in transition.properties:
            trans.set(petri_properties.TRANS_GUARD, transition.properties[petri_properties.TRANS_GUARD])
        if petri_properties.READ_VARIABLE in transition.properties:
            read_variables = transition.properties[petri_properties.READ_VARIABLE]
            for rv in read_variables:
                rv_el = etree.SubElement(trans, petri_properties.READ_VARIABLE)
                rv_el.text = rv
        if petri_properties.WRITE_VARIABLE in transition.properties:
            write_variables = transition.properties[petri_properties.WRITE_VARIABLE]
            for wv in write_variables:
                wv_el = etree.SubElement(trans, petri_properties.WRITE_VARIABLE)
                wv_el.text = wv
    for arc in petrinet.arcs:
        arc_el = etree.SubElement(page, "arc")
        arc_el.set("id", str(hash(arc)))
        if type(arc.source) is PetriNet.Place:
            arc_el.set("source", str(places_map[arc.source]))
            arc_el.set("target", str(transitions_map[arc.target]))
        else:
            arc_el.set("source", str(transitions_map[arc.source]))
            arc_el.set("target", str(places_map[arc.target]))

        if arc.weight > 1:
            inscription = etree.SubElement(arc_el, "inscription")
            arc_weight = etree.SubElement(inscription, "text")
            arc_weight.text = str(arc.weight)

        if isinstance(arc, ResetNet.ResetArc):
            element = etree.SubElement(arc_el, petri_properties.ARCTYPE)
            element_text = etree.SubElement(element, "text")
            element_text.text = petri_properties.RESET_ARC
        elif isinstance(arc, InhibitorNet.InhibitorArc):
            element = etree.SubElement(arc_el, petri_properties.ARCTYPE)
            element_text = etree.SubElement(element, "text")
            element_text.text = petri_properties.INHIBITOR_ARC

        for prop_key in arc.properties:
            if prop_key != petri_properties.ARCTYPE:
                element = etree.SubElement(arc_el, prop_key)
                element_text = etree.SubElement(element, "text")
                element_text.text = str(arc.properties[prop_key])

    if len(final_marking) > 0:
        finalmarkings = etree.SubElement(net, "finalmarkings")
        marking = etree.SubElement(finalmarkings, "marking")

        for place in final_marking:
            placem = etree.SubElement(marking, "place")
            placem.set("idref", place.name)
            placem_text = etree.SubElement(placem, "text")
            placem_text.text = str(final_marking[place])

    # specific for data Petri nets
    if petri_properties.VARIABLES in petrinet.properties:
        variables = etree.SubElement(net, "variables")
        for prop in petrinet.properties[petri_properties.VARIABLES]:
            variable = etree.SubElement(variables, "variable")
            variable.set("type", prop["type"])
            variable_name = etree.SubElement(variable, "name")
            variable_name.text = prop["name"]

    tree = etree.ElementTree(root)

    return tree


def export_petri_as_string(petrinet, marking, final_marking=None, export_prom5=False,
                           parameters=None):
    """
    Parameters
    ----------
    petrinet: :class:`pm4py.entities.petri.petrinet.PetriNet`
        Petri net
    marking: :class:`pm4py.entities.petri.petrinet.Marking`
        Marking
    final_marking: :class:`pm4py.entities.petri.petrinet.Marking`
        Final marking (optional)
    export_prom5
        Enables exporting PNML files in a format that is ProM5-friendly

    Returns
    ----------
    string
        Petri net as string
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(Parameters.ENCODING, parameters, constants.DEFAULT_ENCODING)

    # gets the XML tree
    tree = export_petri_tree(petrinet, marking, final_marking=final_marking,
                             export_prom5=export_prom5)

    # removing default decoding (return binary string as in other parts of the application)
    return etree.tostring(tree, xml_declaration=True, encoding=encoding)


def export_net(petrinet, marking, output_filename, final_marking=None, export_prom5=False,
               parameters=None):
    """
    Export a Petrinet to a PNML file

    Parameters
    ----------
    petrinet: :class:`pm4py.entities.petri.petrinet.PetriNet`
        Petri net
    marking: :class:`pm4py.entities.petri.petrinet.Marking`
        Marking
    final_marking: :class:`pm4py.entities.petri.petrinet.Marking`
        Final marking (optional)
    output_filename:
        Absolute output file name for saving the pnml file
    export_prom5
        Enables exporting PNML files in a format that is ProM5-friendly
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(Parameters.ENCODING, parameters, constants.DEFAULT_ENCODING)

    # gets the XML tree
    tree = export_petri_tree(petrinet, marking, final_marking=final_marking,
                             export_prom5=export_prom5)

    # write the tree to a file
    F = open(output_filename, "wb")
    tree.write(F, pretty_print=True, xml_declaration=True, encoding=encoding)
    F.close()



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
PARAM_EVENT_ID = "param:event:id"
PARAM_EVENT_ACTIVITY = "param:event:activity"
PARAM_EVENT_TIMESTAMP = "param:event:timestamp"
PARAM_OBJECT_ID = "param:object:id"
PARAM_OBJECT_TYPE = "param:object:type"
PARAM_OBJECT_TYPE_PREFIX_EXTENDED = "param:object:type:prefix:extended"
PARAM_QUALIFIER = "param:qualifier"
PARAM_CHNGD_FIELD = "param:chngfield"

DEFAULT_EVENT_ID = "ocel:eid"
DEFAULT_EVENT_ACTIVITY = "ocel:activity"
DEFAULT_EVENT_TIMESTAMP = "ocel:timestamp"
DEFAULT_OBJECT_ID = "ocel:oid"
DEFAULT_OBJECT_TYPE = "ocel:type"
DEFAULT_OBJECT_TYPE_PREFIX_EXTENDED = "ocel:type:"
DEFAULT_QUALIFIER = "ocel:qualifier"
DEFAULT_CHNGD_FIELD = "ocel:field"

OCEL_PREFIX = "ocel:"
OCEL_EVENTS_KEY = "ocel:events"
OCEL_OBJECTS_KEY = "ocel:objects"
OCEL_ID_KEY = "ocel:id"
OCEL_OMAP_KEY = "ocel:omap"
OCEL_TYPED_OMAP_KEY = "ocel:typedOmap"
OCEL_VMAP_KEY = "ocel:vmap"
OCEL_OVMAP_KEY = "ocel:ovmap"
OCEL_O2O_KEY = "ocel:o2o"
OCEL_OBJCHANGES_KEY = "ocel:objectChanges"
OCEL_EVTYPES_KEY = "ocel:eventTypes"
OCEL_OBJTYPES_KEY = "ocel:objectTypes"
OCEL_GLOBAL_LOG = "ocel:global-log"
OCEL_GLOBAL_LOG_ATTRIBUTE_NAMES = "ocel:attribute-names"
OCEL_GLOBAL_LOG_OBJECT_TYPES = "ocel:object-types"
OCEL_GLOBAL_LOG_VERSION = "ocel:version"
OCEL_GLOBAL_LOG_ORDERING = "ocel:ordering"
OCEL_GLOBAL_EVENT = "ocel:global-event"
OCEL_GLOBAL_OBJECT = "ocel:global-object"

PARAM_INTERNAL_INDEX = "param:internal:index"
DEFAULT_INTERNAL_INDEX = "@@index"

DEFAULT_GLOBAL_EVENT = {"ocel:activity": "__INVALID__"}
DEFAULT_GLOBAL_OBJECT = {"ocel:type": "__INVALID__"}
DEFAULT_ORDERING = "timestamp"
CURRENT_VERSION = "1.0"



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from enum import Enum

from pm4py.objects.process_tree.importer.variants import ptml
from pm4py.util import exec_utils


class Variants(Enum):
    PTML = ptml


DEFAULT_VARIANT = Variants.PTML


def apply(file_path, variant=DEFAULT_VARIANT, parameters=None):
    """
    Imports a process tree from the specified path

    Parameters
    ---------------
    path
        Path
    variant
        Variant of the algorithm, possible values:
            - Variants.PTML
    parameters
        Possible parameters (version specific)

    Returns
    ---------------
    tree
        Process tree
    """
    return exec_utils.get_variant(variant).apply(file_path, parameters=parameters)


def deserialize(tree_string, variant=DEFAULT_VARIANT, parameters=None):
    """
    Deserialize a text/binary string representing a process tree in the PTML format

    Parameters
    ----------
    tree_string
        Process tree expressed as PTML string
    variant
        Variant of the algorithm, possible values:
            - Variants.PTML
    parameters
        Other parameters of the algorithm

    Returns
    ----------
    tree
        Process tree
    """
    return exec_utils.get_variant(variant).import_tree_from_string(tree_string, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import Dict, Any, Optional, List

from pm4py.objects.ocel.obj import OCEL
from pm4py.util import pandas_utils


def related_events_dct(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> Dict[str, Dict[str, List[str]]]:
    if parameters is None:
        parameters = {}

    object_types = pandas_utils.format_unique(ocel.relations[ocel.object_type_column].unique())
    dct = {}
    for ot in object_types:
        dct[ot] = ocel.relations[ocel.relations[ocel.object_type_column] == ot].groupby(ocel.object_id_column)[ocel.event_id_column].apply(
            list).to_dict()
    return dct



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import base64
import os
import subprocess
import sys
from typing import Optional, Dict


MAX_EDGE_PENWIDTH_GRAPHVIZ = 2.6
MIN_EDGE_PENWIDTH_GRAPHVIZ = 1.0


def human_readable_stat(timedelta, stat_locale: Optional[Dict[str, str]] = None) -> str:
    """
    Transform a timedelta into a human readable string

    Parameters
    ----------
    timedelta
        Timedelta

    Returns
    ----------
    string
        Human readable string
    """
    if stat_locale is None:
        stat_locale = {}

    c = int(float(timedelta))
    years = c // 31104000
    months = c // 2592000
    days = c // 86400
    hours = c // 3600 % 24
    minutes = c // 60 % 60
    seconds = c % 60
    if years > 0:
        return str(years) + stat_locale.get("year", "Y")
    elif months > 0:
        return str(months) + stat_locale.get("month", "MO")
    elif days > 0:
        return str(days) + stat_locale.get("day", "D")
    elif hours > 0:
        return str(hours) + stat_locale.get("hour", "h")
    elif minutes > 0:
        return str(minutes) + stat_locale.get("minute", "m")
    elif seconds > 0:
        return str(seconds) + stat_locale.get("second", "s")
    else:
        c = int(float(timedelta*1000))
        if c > 0:
            return str(c) + stat_locale.get("millisecond", "ms")
        else:
            return str(int(float(timedelta * 10**9))) + stat_locale.get("nanosecond", "ns")


def get_arc_penwidth(arc_measure, min_arc_measure, max_arc_measure):
    """
    Calculate arc width given the current arc measure value, the minimum arc measure value and the
    maximum arc measure value

    Parameters
    -----------
    arc_measure
        Current arc measure value
    min_arc_measure
        Minimum measure value among all arcs
    max_arc_measure
        Maximum measure value among all arcs

    Returns
    -----------
    penwidth
        Current arc width in the graph
    """
    return MIN_EDGE_PENWIDTH_GRAPHVIZ + (MAX_EDGE_PENWIDTH_GRAPHVIZ - MIN_EDGE_PENWIDTH_GRAPHVIZ) * (
            arc_measure - min_arc_measure) / (max_arc_measure - min_arc_measure + 0.00001)


def get_trans_freq_color(trans_count, min_trans_count, max_trans_count):
    """
    Gets transition frequency color

    Parameters
    ----------
    trans_count
        Current transition count
    min_trans_count
        Minimum transition count
    max_trans_count
        Maximum transition count

    Returns
    ----------
    color
        Frequency color for visible transition
    """
    trans_base_color = int(255 - 100 * (trans_count - min_trans_count) / (max_trans_count - min_trans_count + 0.00001))
    trans_base_color_hex = str(hex(trans_base_color))[2:].upper()
    return "#" + trans_base_color_hex + trans_base_color_hex + "FF"


def get_base64_from_gviz(gviz):
    """
    Get base 64 from string content of the file

    Parameters
    -----------
    gviz
        Graphviz diagram

    Returns
    -----------
    base64
        Base64 string
    """
    render = gviz.render(view=False)
    with open(render, "rb") as f:
        return base64.b64encode(f.read())


def get_base64_from_file(temp_file):
    """
    Get base 64 from string content of the file

    Parameters
    -----------
    temp_file
        Temporary file path

    Returns
    -----------
    base64
        Base64 string
    """
    with open(temp_file, "rb") as f:
        return base64.b64encode(f.read())


def check_visualization_inside_jupyter():
    """
    Checks if the visualization of the model is performed
    inside a Jupyter notebook
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell" or shell == "Shell":
            return True
        else:
            return False
    except NameError:
        return False


def view_image_in_jupyter(file_name):
    """
    Visualizes a picture inside the Jupyter notebooks

    Parameters
    -------------
    file_name
        Name of the file
    """
    from IPython.display import Image
    image = Image(file_name)
    from IPython.display import display
    return display(image)


def open_opsystem_image_viewer(file_name):
    """
    Visualizes a picture using the image viewer of the operating system

    Parameters
    -------------
    file_name
        Name of the file
    """
    if sys.platform.startswith('darwin'):
        subprocess.call(('open', file_name))
    elif os.name == 'nt':  # For Windows
        os.startfile(file_name)
    elif os.name == 'posix':  # For Linux, Mac, etc.
        subprocess.call(('xdg-open', file_name))


def get_corr_hex(num):
    """
    Gets correspondence between a number
    and an hexadecimal string

    Parameters
    -------------
    num
        Number

    Returns
    -------------
    hex_string
        Hexadecimal string
    """
    if num < 10:
        return str(int(num))
    elif num < 11:
        return "A"
    elif num < 12:
        return "B"
    elif num < 13:
        return "C"
    elif num < 14:
        return "D"
    elif num < 15:
        return "E"
    elif num < 16:
        return "F"



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.visualization.transition_system.util import visualize_graphviz
from enum import Enum
from typing import Optional, Dict, Any, Union
from pm4py.objects.transition_system.obj import TransitionSystem
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    SHOW_LABELS = "show_labels"
    SHOW_NAMES = "show_names"
    FORCE_NAMES = "force_names"
    FILLCOLORS = "fillcolors"
    FONT_SIZE = "font_size"


def apply(tsys: TransitionSystem, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> graphviz.Digraph:
    """
    Get visualization of a Transition System

    Parameters
    -----------
    tsys
        Transition system
    parameters
        Optional parameters of the algorithm

    Returns
    ----------
    gviz
        Graph visualization
    """

    gviz = visualize_graphviz.visualize(tsys, parameters=parameters)
    return gviz



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import sys


def indent_representation(model_string, max_indent=sys.maxsize):
    model_string = model_string.replace('\n', '').replace('\r', '').replace('\t', '').strip()

    indent_level = 0
    list_strs = []
    formatted_str = ""
    for char in model_string:
        if char in '({':
            formatted_str += char
            indent_level += 1
            if indent_level <= max_indent:
                list_strs.append(formatted_str)
                formatted_str = '\t' * indent_level
        elif char in ')}':
            indent_level -= 1
            if indent_level < max_indent:
                if formatted_str[-1] not in '({':
                    list_strs.append(formatted_str)
                    formatted_str = '\t' * indent_level
            formatted_str += char
        elif char == ',':
            formatted_str += char
            if indent_level <= max_indent:
                list_strs.append(formatted_str)
                formatted_str = '\t' * indent_level
        else:
            formatted_str += char
    list_strs.append(formatted_str)
    return list_strs



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

EXISTENCE = "existence"
EXACTLY_ONE = "exactly_one"
INIT = "init"
RESPONDED_EXISTENCE = "responded_existence"
RESPONSE = "response"
PRECEDENCE = "precedence"
SUCCESSION = "succession"
ALTRESPONSE = "altresponse"
ALTPRECEDENCE = "altprecedence"
ALTSUCCESSION = "altsuccession"
CHAINRESPONSE = "chainresponse"
CHAINPRECEDENCE = "chainprecedence"
CHAINSUCCESSION = "chainsuccession"
ABSENCE = "absence"
COEXISTENCE = "coexistence"
NONCOEXISTENCE = "noncoexistence"
NONSUCCESSION = "nonsuccession"
NONCHAINSUCCESSION = "nonchainsuccession"



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.algo.organizational_mining.roles.common import algorithm
from pm4py.util import xes_constants as xes
from collections import Counter
from pm4py.util import exec_utils

from enum import Enum
from pm4py.util import constants
from typing import Optional, Dict, Any, Union, List
import pandas as pd
from pm4py.objects.org.roles.obj import Role


class Parameters(Enum):
    ROLES_THRESHOLD_PARAMETER = "roles_threshold_parameter"
    RESOURCE_KEY = constants.PARAMETER_CONSTANT_RESOURCE_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def apply(df: pd.DataFrame, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> List[Role]:
    """
    Gets the roles (group of different activities done by similar resources)
    out of the log

    Parameters
    -------------
    df
        Pandas dataframe
    parameters
        Possible parameters of the algorithm

    Returns
    ------------
    roles
        List of different roles inside the log
    """
    if parameters is None:
        parameters = {}

    resource_key = exec_utils.get_param_value(Parameters.RESOURCE_KEY, parameters, xes.DEFAULT_RESOURCE_KEY)
    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY)
    activity_resource_couples = Counter(df.groupby([resource_key, activity_key]).size().to_dict())

    return algorithm.apply(activity_resource_couples, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.algo.conformance.log_skeleton.variants import classic
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List, Set
from pm4py.objects.log.obj import EventLog, Trace
import pandas as pd


class Variants(Enum):
    CLASSIC = classic


CLASSIC = Variants.CLASSIC
DEFAULT_VARIANT = Variants.CLASSIC


def apply(obj: Union[EventLog, Trace, pd.DataFrame], model: Dict[str, Any], variant=DEFAULT_VARIANT, parameters: Optional[Dict[Any, Any]] = None) -> List[Set[Any]]:
    """
    Apply log-skeleton based conformance checking given an event log/trace
    and a log-skeleton model

    Parameters
    --------------
    obj
        Object (event log/trace)
    model
        Log-skeleton model
    variant
        Variant of the algorithm, possible values: Variants.CLASSIC
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.CONSIDERED_CONSTRAINTS, among: equivalence, always_after, always_before, never_together, directly_follows, activ_freq

    Returns
    --------------
    aligned_traces
        Conformance checking results for each trace:
        - Outputs.IS_FIT => boolean that tells if the trace is perfectly fit according to the model
        - Outputs.DEV_FITNESS => deviation based fitness (between 0 and 1; the more the trace is near to 1 the more fit is)
        - Outputs.DEVIATIONS => list of deviations in the model
    """
    if parameters is None:
        parameters = {}

    if type(obj) is Trace:
        return exec_utils.get_variant(variant).apply_trace(obj, model, parameters=parameters)
    else:
        return exec_utils.get_variant(variant).apply_log(obj, model, parameters=parameters)


def apply_from_variants_list(var_list: List[List[str]], model: Dict[str, Any], variant=DEFAULT_VARIANT, parameters: Optional[Dict[Any, Any]] = None) -> List[Set[Any]]:
    """
    Performs conformance checking using the log skeleton,
    applying it from a list of variants

    Parameters
    --------------
    var_list
        List of variants
    model
        Log skeleton model
    variant
        Variant of the algorithm, possible values: Variants.CLASSIC
    parameters
        Parameters

    Returns
    --------------
    conformance_dictio
        Dictionary containing, for each variant, the result
        of log skeleton checking
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply_from_variants_list(var_list, model, parameters=parameters)


def get_diagnostics_dataframe(log: EventLog, conf_result: List[Set[Any]], variant=DEFAULT_VARIANT, parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Gets the diagnostics dataframe from a log and the results
    of log skeleton-based conformance checking

    Parameters
    --------------
    log
        Event log
    conf_result
        Results of conformance checking

    Returns
    --------------
    diagn_dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).get_diagnostics_dataframe(log, conf_result, parameters=parameters)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import math
from datetime import datetime
from enum import Enum
from typing import Optional, Dict, Any, Tuple, List

import numpy as np
from pm4py.util import ml_utils

from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventLog
from pm4py.algo.transformation.log_to_features import algorithm as log_to_features
from pm4py.objects.log.util import sorting
from pm4py.util import constants, xes_constants
from pm4py.util import exec_utils, pandas_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def smooth(y: np.ndarray, box_pts: int) -> np.ndarray:
    """
    Smooths the points in y with a weighted average.

    Parameters
    ----------------
    y
        Points
    box_pts
        Size of the weighted average

    Returns
    ----------------
    y_smooth
        Smoothened y
    """
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode='same')
    return y_smooth


def apply(log: EventLog, parameters: Optional[Dict[str, Any]] = None) -> Tuple[List[datetime], np.ndarray]:
    """
    Analyse the evolution of the features over the time using a locally linear embedding.

    Parameters
    -----------------
    log
        Event log
    parameters
        Variant-specific parameters, including:
        - Parameters.ACTIVITY_KEY => the activity key
        - Parameters.TIMESTAMP_KEY => the timestamp key
        - Parameters.CASE_ID_KEY => the case ID key

    Returns
    ----------------
    x
        Date attributes (starting points of the cases)
    y
        Deviation from the standard behavior (higher absolute values of y signal a higher deviation
        from the standard behavior)
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)

    if pandas_utils.check_is_pandas_dataframe(log):
        # keep only the needed columns
        case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
        log = log[[case_id_key, activity_key, timestamp_key]]

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters)
    log = sorting.sort_timestamp(log, timestamp_key)

    x = [trace[0][timestamp_key] for trace in log]
    data, feature_names = log_to_features.apply(log, parameters={"str_ev_attr": [activity_key], "str_evsucc_attr": [activity_key]})
    data = np.array([np.array(x) for x in data])

    tsne = ml_utils.LocallyLinearEmbedding(n_components=1, eigen_solver='dense')
    data = tsne.fit_transform(data)
    data = np.ndarray.flatten(data)

    y = data
    smooth_amount = 1 + math.floor(math.sqrt(len(y)))
    y = smooth(y, smooth_amount)

    return x, y



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import base64
import os
import re
import shutil
import tempfile
import webbrowser

from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from pm4py.visualization.powl.variants import basic
from pm4py.visualization.powl.variants import net
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any
from pm4py.objects.powl.obj import POWL
import graphviz


class POWLVisualizationVariants(Enum):
    BASIC = basic
    NET = net


class Parameters(Enum):
    FORMAT = "format"
    ENCODING = "encoding"


DEFAULT_VARIANT = POWLVisualizationVariants.BASIC


def apply(powl: POWL, variant=DEFAULT_VARIANT, frequency_tags=True, parameters: Optional[Dict[Any, Any]] = None)\
        -> str:
    """
    Method for POWL model representation

    Parameters
    -----------
    powl
        POWL model
    parameters
        Possible parameters of the algorithm:
            Parameters.FORMAT -> Format of the image (PDF, PNG, SVG; default PNG)
    variant
        Variant of the algorithm to use:
            - POWLVisualizationVariants.BASIC (default)
            - POWLVisualizationVariants.NET: BPMN-like visualization with decision gates
    frequency_tags
        Simplify the visualization using frequency tags

    Returns
    -----------
    str
        SVG Content
    """
    if parameters is None:
        parameters = {}

    if frequency_tags:
        powl = powl.simplify_using_frequent_transitions()

    viz = exec_utils.get_variant(variant).apply(powl)
    svg_content = viz.pipe().decode('utf-8')

    def inline_images_and_svgs(svg_content):
        img_pattern = re.compile(r'<image[^>]+xlink:href=["\'](.*?)["\'][^>]*>')

        def encode_file_to_base64(file_path):
            with open(file_path, 'rb') as file:
                return base64.b64encode(file.read()).decode('utf-8')

        def read_file_content_and_viewbox(file_path):
            with open(file_path, 'r') as file:
                content = file.read()
                content = re.sub(r'<\?xml.*?\?>', '', content, flags=re.DOTALL)
                content = re.sub(r'<!DOCTYPE.*?>', '', content, flags=re.DOTALL)
                viewBox_match = re.search(r'viewBox="([^"]*)"', content)
                viewBox = viewBox_match.group(1) if viewBox_match else "0 0 1 1"
                svg_content_match = re.search(r'<svg[^>]*>(.*?)</svg>', content, re.DOTALL)
                svg_content = svg_content_match.group(1) if svg_content_match else content
                return svg_content, viewBox

        def replace_with_inline_content(match):
            file_path = match.group(1)
            if file_path.lower().endswith('.svg'):
                svg_data, viewBox = read_file_content_and_viewbox(file_path)
                viewBox_values = [float(v) for v in viewBox.split()]
                actual_width, actual_height = viewBox_values[2], viewBox_values[3]

                intended_width = float(match.group(0).split('width="')[1].split('"')[0].replace('px', ''))
                intended_height = float(match.group(0).split('height="')[1].split('"')[0].replace('px', ''))
                x = float(match.group(0).split('x="')[1].split('"')[0])
                y = float(match.group(0).split('y="')[1].split('"')[0])

                scale_x = intended_width / actual_width
                scale_y = intended_height / actual_height

                return f'<g transform="translate({x},{y}) scale({scale_x},{scale_y})">{svg_data}</g>'
            else:
                base64_data = encode_file_to_base64(file_path)
                return match.group(0).replace(file_path, f"data:image/png;base64,{base64_data}")

        return img_pattern.sub(replace_with_inline_content, svg_content)

    svg_content_with_inline_images = inline_images_and_svgs(svg_content)

    return svg_content_with_inline_images


def save(svg_content: str, output_file_path: str, parameters: Optional[Dict[Any, Any]] = None):
    """
    Save the diagram in a specified format.

    Parameters
    -----------
    svg_content : str
        SVG content
    output_file_path : str
        Path where the output file should be saved
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(Parameters.ENCODING, parameters, "utf-8")

    with tempfile.NamedTemporaryFile(delete=False, mode='w+', suffix='.svg') as tmpfile:
        tmpfile.write(svg_content)
        tmpfile_path = tmpfile.name

    if output_file_path.endswith("svg"):
        shutil.move(tmpfile_path, output_file_path)
    elif output_file_path.endswith("png"):
        import cairosvg as cairosvg
        cairosvg.svg2png(bytestring=svg_content.encode(encoding), write_to=output_file_path)
    elif output_file_path.endswith("pdf"):
        import cairosvg as cairosvg
        cairosvg.svg2pdf(bytestring=svg_content.encode(encoding), write_to=output_file_path)
    else:
        raise Exception(f"Unsupported format! Please use 'svg', 'png', or 'pdf'.")

    if os.path.exists(tmpfile_path):
        os.remove(tmpfile_path)


def view(svg_content: str, parameters: Optional[Dict[Any, Any]] = None):
    """
    View the diagram

    Parameters
    -----------
    svg_content
        SVG content
    image_format
        image format
    """
    if parameters is None:
        parameters = {}

    image_format = str(exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")).lower()
    encoding = exec_utils.get_param_value(Parameters.ENCODING, parameters, "utf-8")

    with tempfile.NamedTemporaryFile(delete=False, mode='w+', suffix='.svg') as tmpfile:
        tmpfile.write(svg_content)
        tmpfile_path = tmpfile.name

        if image_format == "svg":
            absolute_path = os.path.abspath(tmpfile_path)
            return webbrowser.open('file://' + absolute_path)
        elif image_format == "png":
            import cairosvg as cairosvg
            cairosvg.svg2png(bytestring=svg_content.encode(encoding), url=tmpfile_path, write_to=tmpfile_path + '.png')
            webbrowser.open('file://' + os.path.abspath(tmpfile_path + '.png'))
        elif image_format == "pdf":
            import cairosvg as cairosvg
            cairosvg.svg2pdf(bytestring=svg_content.encode(encoding), url=tmpfile_path, write_to=tmpfile_path + '.pdf')
            webbrowser.open('file://' + os.path.abspath(tmpfile_path + '.pdf'))
        else:
            raise Exception(f"Unsupported format: {image_format}. Please use 'svg', 'png', or 'pdf'.")



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from typing import Optional, Dict, Any
from graphviz import Digraph
from enum import Enum
from pm4py.util import exec_utils, constants
import tempfile
from uuid import uuid4
from pm4py.util import vis_utils
from statistics import mean, median


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    ACT_METRIC = "act_metric"
    EDGE_METRIC = "edge_metric"
    ACT_THRESHOLD = "act_threshold"
    EDGE_THRESHOLD = "edge_threshold"
    ANNOTATION = "annotation"
    PERFORMANCE_AGGREGATION_MEASURE = "aggregationMeasure"


def ot_to_color(ot: str) -> str:
    ot = int(hash(ot))
    num = []
    while len(num) < 6:
        num.insert(0, ot % 16)
        ot = ot // 16
    ret = "#" + "".join([vis_utils.get_corr_hex(x) for x in num])
    return ret


def add_activity(G: Digraph, act, freq, act_prefix, nodes, annotation, min_freq, max_freq):
    """
    Adds an activity node to the graph
    """
    act_uuid = str(uuid4())
    nodes[act] = act_uuid
    fillcolor = vis_utils.get_trans_freq_color(freq, min_freq, max_freq)
    if annotation == "frequency":
        G.node(act_uuid, label=act + "\n" + act_prefix + str(freq), shape="box", style="filled", fillcolor=fillcolor)
    else:
        G.node(act_uuid, label=act, shape="box")


def add_frequency_edge(G: Digraph, ot, act1, act2, freq, edge_prefix, nodes, min_freq, max_freq):
    """
    Adds a edge (frequency annotation)
    """
    otc = ot_to_color(ot)
    act_uuid1 = nodes[act1]
    act_uuid2 = nodes[act2]
    penwidth = vis_utils.get_arc_penwidth(freq, min_freq, max_freq)
    G.edge(act_uuid1, act_uuid2, label=ot + " " + edge_prefix + str(freq), fontsize="8", penwidth=str(penwidth),
           color=otc, fontcolor=otc)


def add_performance_edge(G: Digraph, ot, act1, act2, perf, edge_prefix, nodes, aggregation_measure):
    """
    Adds an edge (performance annotation)
    """
    otc = ot_to_color(ot)
    if aggregation_measure == "median":
        perf = median(perf)
    elif aggregation_measure == "min":
        perf = min(perf)
    elif aggregation_measure == "max":
        perf = max(perf)
    elif aggregation_measure == "sum":
        perf = sum(perf)
    else:
        perf = mean(perf)
    act_uuid1 = nodes[act1]
    act_uuid2 = nodes[act2]
    G.edge(act_uuid1, act_uuid2, label=ot + " " + edge_prefix + vis_utils.human_readable_stat(perf), fontsize="8",
           color=otc, fontcolor=otc)


def add_start_node(G: Digraph, ot, act, freq, edge_prefix, nodes, annotation, min_freq, max_freq):
    """
    Adds a start node to the graph
    """
    otc = ot_to_color(ot)
    act_uuid = nodes[act]
    start_ot = "start_node@#@#" + ot
    if start_ot not in nodes:
        endpoint_uuid = str(uuid4())
        nodes[start_ot] = endpoint_uuid
        G.node(endpoint_uuid, label=ot, shape="ellipse", style="filled", fillcolor=otc)
    start_ot_uuid = nodes[start_ot]
    edge_label = ""
    if annotation == "frequency":
        edge_label = ot + " " + edge_prefix + str(freq)
    penwidth = vis_utils.get_arc_penwidth(freq, min_freq, max_freq)
    G.edge(start_ot_uuid, act_uuid, label=edge_label, fontsize="8", penwidth=str(penwidth), fontcolor=otc, color=otc)


def add_end_node(G: Digraph, ot, act, freq, edge_prefix, nodes, annotation, min_freq, max_freq):
    """
    Adds an end node to the graph
    """
    otc = ot_to_color(ot)
    act_uuid = nodes[act]
    end_ot = "end_node@#@#" + ot
    if end_ot not in nodes:
        endpoint_uuid = str(uuid4())
        nodes[end_ot] = endpoint_uuid
        G.node(endpoint_uuid, label=ot, shape="underline", fontcolor=otc)
    end_ot_uuid = nodes[end_ot]
    edge_label = ""
    if annotation == "frequency":
        edge_label = ot + " " + edge_prefix + str(freq)
    penwidth = vis_utils.get_arc_penwidth(freq, min_freq, max_freq)
    G.edge(act_uuid, end_ot_uuid, label=edge_label, fontsize="8", penwidth=str(penwidth), fontcolor=otc, color=otc)


def apply(ocdfg: Dict[str, Any], parameters: Optional[Dict[Any, Any]] = None) -> Digraph:
    """
    Visualizes an OC-DFG as a Graphviz di-graph

    Parameters
    ---------------
    ocdfg
        OC-DFG
    parameters
        Parameters of the algorithm:
        - Parameters.FORMAT => the format of the output visualization (default: "png")
        - Parameters.BGCOLOR => the default background color (default: "bgcolor")
        - Parameters.RANKDIR => direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)
        - Parameters.ACT_METRIC => the metric to use for the activities. Available values:
            - "events" => number of events (default)
            - "unique_objects" => number of unique objects
            - "total_objects" => number of total objects
        - Parameters.EDGE_METRIC => the metric to use for the edges. Available values:
            - "event_couples" => number of event couples (default)
            - "unique_objects" => number of unique objects
            - "total_objects" => number of total objects
        - Parameters.ACT_THRESHOLD => the threshold to apply on the activities frequency (default: 0). Only activities
        having a frequency >= than this are kept in the graph.
        - Parameters.EDGE_THRESHOLD => the threshold to apply on the edges frequency (default 0). Only edges
        having a frequency >= than this are kept in the graph.
        - Parameters.ANNOTATION => the annotation to use for the visualization. Values:
            - "frequency": frequency annotation
            - "performance": performance annotation
        - Parameters.PERFORMANCE_AGGREGATION_MEASURE => the aggregation measure to use for the performance:
            - mean
            - median
            - min
            - max
            - sum

    Returns
    ---------------
    viz
        Graphviz DiGraph
    """
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR)
    rankdir = exec_utils.get_param_value(Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ)
    act_metric = exec_utils.get_param_value(Parameters.ACT_METRIC, parameters, "events")
    edge_metric = exec_utils.get_param_value(Parameters.EDGE_METRIC, parameters, "event_couples")
    act_threshold = exec_utils.get_param_value(Parameters.ACT_THRESHOLD, parameters, 0)
    edge_threshold = exec_utils.get_param_value(Parameters.EDGE_THRESHOLD, parameters, 0)
    annotation = exec_utils.get_param_value(Parameters.ANNOTATION, parameters, "frequency")
    performance_aggregation_measure = exec_utils.get_param_value(Parameters.PERFORMANCE_AGGREGATION_MEASURE, parameters,
                                                                 "mean")

    act_count = {}
    act_ot_count = {}
    sa_count = {}
    ea_count = {}
    act_prefix = ""
    edges_count = {}
    edges_performance = {}
    edge_prefix = ""

    if act_metric == "events":
        act_count = ocdfg["activities_indep"]["events"]
        act_ot_count = ocdfg["activities_ot"]["events"]
        sa_count = ocdfg["start_activities"]["events"]
        ea_count = ocdfg["end_activities"]["events"]
        act_prefix = "E="
    elif act_metric == "unique_objects":
        act_count = ocdfg["activities_indep"]["unique_objects"]
        act_ot_count = ocdfg["activities_ot"]["unique_objects"]
        sa_count = ocdfg["start_activities"]["unique_objects"]
        ea_count = ocdfg["end_activities"]["unique_objects"]
        act_prefix = "UO="
    elif act_metric == "total_objects":
        act_count = ocdfg["activities_indep"]["total_objects"]
        act_ot_count = ocdfg["activities_ot"]["total_objects"]
        sa_count = ocdfg["start_activities"]["total_objects"]
        ea_count = ocdfg["end_activities"]["total_objects"]
        act_prefix = "TO="

    if edge_metric == "event_couples":
        edges_count = ocdfg["edges"]["event_couples"]
        edges_performance = ocdfg["edges_performance"]["event_couples"]
        edge_prefix = "EC="
    elif edge_metric == "unique_objects":
        edges_count = ocdfg["edges"]["unique_objects"]
        edge_prefix = "UO="
    elif edge_metric == "total_objects":
        edges_count = ocdfg["edges"]["total_objects"]
        edges_performance = ocdfg["edges_performance"]["total_objects"]
        edge_prefix = "TO="

    if annotation == "performance" and edge_metric == "unique_objects":
        raise Exception("unsupported performance visualization for unique objects!")

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    filename.close()

    viz = Digraph("ocdfg", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})
    viz.attr('node', shape='ellipse', fixedsize='false')

    min_edges_count = {}
    max_edges_count = {}

    for ot in edges_count:
        all_edges_count = [len(y) for y in edges_count[ot].values()]
        min_edges_count[ot] = min(all_edges_count)
        max_edges_count[ot] = max(all_edges_count)
        all_sa_count = [len(y) for y in sa_count[ot].values()]
        min_edges_count[ot] = min(min(all_sa_count), min_edges_count[ot])
        max_edges_count[ot] = max(max(all_sa_count), max_edges_count[ot])
        all_ea_count = [len(y) for y in ea_count[ot].values()]
        min_edges_count[ot] = min(min(all_ea_count), min_edges_count[ot])
        max_edges_count[ot] = max(max(all_ea_count), max_edges_count[ot])

    act_count_values = [len(y) for y in act_count.values()]
    min_act_count = min(act_count_values)
    max_act_count = max(act_count_values)

    nodes = {}
    for act in act_count:
        if len(act_count[act]) >= act_threshold:
            add_activity(viz, act, len(act_count[act]), act_prefix, nodes, annotation, min_act_count, max_act_count)

    for ot in edges_count:
        for act_cou in edges_count[ot]:
            if act_cou[0] in nodes and act_cou[1] in nodes:
                if len(edges_count[ot][act_cou]) >= edge_threshold:
                    if annotation == "frequency":
                        add_frequency_edge(viz, ot, act_cou[0], act_cou[1], len(edges_count[ot][act_cou]), edge_prefix,
                                           nodes, min_edges_count[ot], max_edges_count[ot])
                    elif annotation == "performance":
                        add_performance_edge(viz, ot, act_cou[0], act_cou[1], edges_performance[ot][act_cou],
                                             edge_prefix, nodes, performance_aggregation_measure)

    for ot in sa_count:
        for act in sa_count[ot]:
            if act in nodes:
                if len(sa_count[ot][act]) >= edge_threshold:
                    miec = min_edges_count[ot] if ot in min_edges_count else len(sa_count[ot][act])
                    maec = max_edges_count[ot] if ot in max_edges_count else len(sa_count[ot][act])
                    add_start_node(viz, ot, act, len(sa_count[ot][act]), edge_prefix, nodes, annotation,
                                   miec, maec)

    for ot in ea_count:
        for act in ea_count[ot]:
            if act in nodes:
                if len(ea_count[ot][act]) >= edge_threshold:
                    miec = min_edges_count[ot] if ot in min_edges_count else len(ea_count[ot][act])
                    maec = max_edges_count[ot] if ot in max_edges_count else len(ea_count[ot][act])
                    add_end_node(viz, ot, act, len(ea_count[ot][act]), edge_prefix, nodes, annotation,
                                 miec, maec)

    viz.attr(rankdir=rankdir)
    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

from typing import Optional, Dict, Any
import pandas as pd
from datetime import datetime
from pm4py.util import pandas_utils
import importlib.util


def apply(parameters: Optional[Dict[str, Any]] = None) -> pd.DataFrame:
    """
    Extract a process mining dataframe from all the events recorded in the Windows registry.

    CASE ID (case:concept:name) => name of the computer emitting the events.
    ACTIVITY (concept:name)  => concatenation of the source name of the event and the event identifier
                (see https://learn.microsoft.com/en-us/previous-versions/windows/desktop/eventlogprov/win32-ntlogevent)
    TIMESTAMP (time:timestamp) => timestamp of generation of the event
    RESOURCE (org:resource) => username involved in the event

    Returns
    ----------------
    dataframe
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    import win32com.client

    print(":: executing SQL query against the Windows registry. this can take time.")

    strComputer = "."
    objWMIService = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    objSWbemServices = objWMIService.ConnectServer(strComputer, "root\cimv2")
    colItems = objSWbemServices.ExecQuery("Select * from Win32_NTLogEvent")
    events = []

    progress = None
    if importlib.util.find_spec("tqdm"):
        from tqdm.auto import tqdm
        progress = tqdm(total=len(colItems),
                        desc="extracting Windows events, progress :: ")

    for objItem in colItems:
        events.append({"category": str(objItem.Properties_("Category")), "categoryString": str(objItem.Properties_("CategoryString")),
                       "computerName": str(objItem.Properties_("ComputerName")), "eventCode": str(objItem.Properties_("EventCode")),
                       "eventIdentifier": str(objItem.Properties_("EventIdentifier")), "eventType": str(objItem.Properties_("EventType")),
                       "logFile": str(objItem.Properties_("LogFile")), "message": str(objItem.Properties_("Message")),
                       "recordNumber": str(objItem.Properties_("RecordNumber")),
                       "sourceName": str(objItem.Properties_("SourceName")),
                       "timeGenerated": datetime.strptime(str(str(objItem.Properties_("TimeGenerated"))).split("+")[0].split("-")[0], "%Y%m%d%H%M%S.%f"),
                       "timeWritten": datetime.strptime(str(str(objItem.Properties_("TimeWritten"))).split("+")[0].split("-")[0], "%Y%m%d%H%M%S.%f"),
                       "type": str(str(objItem.Properties_("Type"))), "user": str(str(objItem.Properties_("User")))})
        if progress is not None:
            progress.update()

    if progress is not None:
        progress.close()

    dataframe = pandas_utils.instantiate_dataframe(events)
    dataframe["case:concept:name"] = dataframe["computerName"]
    dataframe["time:timestamp"] = dataframe["timeGenerated"]
    dataframe["concept:name"] = dataframe["sourceName"] + " " + dataframe["eventIdentifier"]
    dataframe["org:resource"] = dataframe["user"]
    dataframe = pandas_utils.insert_index(dataframe, "@@index", copy_dataframe=False, reset_index=False)
    dataframe = dataframe.sort_values(["time:timestamp", "@@index"])
    dataframe["@@case_index"] = dataframe.groupby("case:concept:name", sort=False).ngroup()
    dataframe = dataframe.sort_values(["@@case_index", "time:timestamp", "@@index"])

    return dataframe



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.bpmn.obj import BPMN


def reduce_xor_gateways(bpmn_graph, parameters=None):
    """
    Reduces the number of XOR gateways in the diagram

    Parameters
    ------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters

    Returns
    ------------
    bpmn_graph
        (possibly reduced) BPMN graph
    """
    if parameters is None:
        parameters = {}

    changed = True
    while changed:
        changed = False
        outgoing_edges = None
        incoming_edges = None
        outgoing_edges = {}
        incoming_edges = {}

        for flow in bpmn_graph.get_flows():
            source = flow.get_source()
            target = flow.get_target()

            if source not in outgoing_edges:
                outgoing_edges[source] = set()
            outgoing_edges[source].add(flow)

            if target not in incoming_edges:
                incoming_edges[target] = set()
            incoming_edges[target].add(flow)

        nodes = list(bpmn_graph.get_nodes())
        for node in nodes:
            if isinstance(node, BPMN.ExclusiveGateway):
                if node in outgoing_edges and node in incoming_edges and len(outgoing_edges[node]) == 1 and len(
                        incoming_edges[node]) == 1:
                    changed = True
                    source_node = None
                    target_node = None
                    for flow in incoming_edges[node]:
                        source_node = flow.get_source()
                        if flow in bpmn_graph.get_flows():
                            bpmn_graph.remove_flow(flow)
                    for flow in outgoing_edges[node]:
                        target_node = flow.get_target()
                        if flow in bpmn_graph.get_flows():
                            bpmn_graph.remove_flow(flow)
                    if node in bpmn_graph.get_nodes():
                        bpmn_graph.remove_node(node)
                    bpmn_graph.add_flow(BPMN.SequenceFlow(source_node, target_node))
                    break

    return bpmn_graph


def apply(bpmn_graph, parameters=None):
    """
    Reduce the complexity of a BPMN graph by removing useless elements

    Parameters
    ------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters

    Returns
    ------------
    bpmn_graph
        (possibly reduced) BPMN graph
    """
    if parameters is None:
        parameters = {}

    bpmn_graph = reduce_xor_gateways(bpmn_graph, parameters=parameters)
    return bpmn_graph



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''

from enum import Enum
from pm4py.objects.log.obj import EventLog
import pandas as pd
from typing import Union, Dict, Optional, Any, List
from pm4py.algo.discovery.declare.templates import *
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils
from collections import Counter


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def __check_existence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                      parameters: Optional[Dict[Any, Any]] = None):
    if EXISTENCE in model:
        for act in model[EXISTENCE]:
            if act not in trace:
                trace_dict["deviations"].append([EXISTENCE, act])


def __check_absence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                    parameters: Optional[Dict[Any, Any]] = None):
    if ABSENCE in model:
        for act in model[ABSENCE]:
            if act in trace:
                trace_dict["deviations"].append([ABSENCE, act])


def __check_exactly_one(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                        parameters: Optional[Dict[Any, Any]] = None):
    if EXACTLY_ONE in model:
        trace_counter = Counter(trace)
        for act in model[EXACTLY_ONE]:
            if trace_counter[act] != 1:
                trace_dict["deviations"].append([EXACTLY_ONE, act])


def __check_init(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                 parameters: Optional[Dict[Any, Any]] = None):
    if INIT in model:
        for act in model[INIT]:
            if len(trace) == 0 or trace[0] != act:
                trace_dict["deviations"].append([INIT, act])


def __check_responded_existence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                                trace_dict: Dict[str, List[Any]], parameters: Optional[Dict[Any, Any]] = None):
    if RESPONDED_EXISTENCE in model:
        for act_couple in model[RESPONDED_EXISTENCE]:
            if act_couple[0] in trace and act_couple[1] not in trace:
                trace_dict["deviations"].append([RESPONDED_EXISTENCE, act_couple])


def __check_coexistence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                        parameters: Optional[Dict[Any, Any]] = None):
    if COEXISTENCE in model:
        for act_couple in model[COEXISTENCE]:
            if (act_couple[0] in trace and act_couple[1] not in trace) or (
                    act_couple[1] in trace and act_couple[0] not in trace):
                trace_dict["deviations"].append([COEXISTENCE, act_couple])


def __check_non_coexistence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                            trace_dict: Dict[str, List[Any]], parameters: Optional[Dict[Any, Any]] = None):
    if NONCOEXISTENCE in model:
        for act_couple in model[NONCOEXISTENCE]:
            if act_couple[0] in trace and act_couple[1] in trace:
                trace_dict["deviations"].append([NONCOEXISTENCE, act_couple])


def __check_response(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                     act_idxs: Dict[str, List[int]], parameters: Optional[Dict[Any, Any]] = None):
    if RESPONSE in model:
        for act_couple in model[RESPONSE]:
            if act_couple[0] in trace:
                if (not act_couple[1] in trace) or max(act_idxs[act_couple[0]]) > max(act_idxs[act_couple[1]]):
                    trace_dict["deviations"].append([RESPONSE, act_couple])


def __check_precedence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                       act_idxs: Dict[str, List[int]], parameters: Optional[Dict[Any, Any]] = None):
    if PRECEDENCE in model:
        for act_couple in model[PRECEDENCE]:
            if act_couple[1] in trace:
                if (not act_couple[0] in trace) or min(act_idxs[act_couple[0]]) > min(act_idxs[act_couple[1]]):
                    trace_dict["deviations"].append([PRECEDENCE, act_couple])


def __check_succession(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]], trace_dict: Dict[str, List[Any]],
                       act_idxs: Dict[str, List[int]], parameters: Optional[Dict[Any, Any]] = None):
    if SUCCESSION in model:
        for act_couple in model[SUCCESSION]:
            if (not act_couple[0] in trace or not act_couple[1] in trace) or min(act_idxs[act_couple[0]]) > min(
                    act_idxs[act_couple[1]]) or max(act_idxs[act_couple[0]]) > max(act_idxs[act_couple[1]]):
                trace_dict["deviations"].append([SUCCESSION, act_couple])


def __check_alt_response(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                         trace_dict: Dict[str, List[Any]], act_idxs: Dict[str, List[int]],
                         parameters: Optional[Dict[Any, Any]] = None):
    if ALTRESPONSE in model:
        for act_couple in model[RESPONSE]:
            spec_idxs = []
            if act_couple[0] in trace:
                spec_idxs = spec_idxs + [(act_couple[0], i) for i in act_idxs[act_couple[0]]]
            if act_couple[1] in trace:
                spec_idxs = spec_idxs + [(act_couple[1], i) for i in act_idxs[act_couple[1]]]
            spec_idxs = sorted(spec_idxs, key=lambda x: (x[1], x[0]))
            while spec_idxs:
                if spec_idxs[0][0] != act_couple[0]:
                    del spec_idxs[0]
                else:
                    break
            is_ok = True
            for i in range(len(spec_idxs)):
                if i % 2 == 0 and (spec_idxs[i][0] != act_couple[0] or i == len(spec_idxs) - 1 or spec_idxs[i + 1][0] !=
                                   act_couple[1]):
                    is_ok = False
                    break
            if not is_ok:
                trace_dict["deviations"].append([ALTRESPONSE, act_couple])


def __check_chain_response(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                           trace_dict: Dict[str, List[Any]], act_idxs: Dict[str, List[int]],
                           parameters: Optional[Dict[Any, Any]] = None):
    if CHAINRESPONSE in model:
        for act_couple in model[CHAINRESPONSE]:
            spec_idxs = []
            if act_couple[0] in trace:
                spec_idxs = spec_idxs + [(act_couple[0], i) for i in act_idxs[act_couple[0]]]
            if act_couple[1] in trace:
                spec_idxs = spec_idxs + [(act_couple[1], i) for i in act_idxs[act_couple[1]]]
            spec_idxs = sorted(spec_idxs, key=lambda x: (x[1], x[0]))
            while spec_idxs:
                if spec_idxs[0][0] != act_couple[0]:
                    del spec_idxs[0]
                else:
                    break
            is_ok = True
            for i in range(len(spec_idxs)):
                if i % 2 == 0 and (spec_idxs[i][0] != act_couple[0] or i == len(spec_idxs) - 1 or spec_idxs[i + 1][0] !=
                                   act_couple[1] or spec_idxs[i + 1][1] != spec_idxs[i][1] + 1):
                    is_ok = False
                    break
            if not is_ok:
                trace_dict["deviations"].append([CHAINRESPONSE, act_couple])


def __check_alt_precedence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                           trace_dict: Dict[str, List[Any]], act_idxs: Dict[str, List[int]],
                           parameters: Optional[Dict[Any, Any]] = None):
    if ALTPRECEDENCE in model:
        for act_couple in model[ALTPRECEDENCE]:
            spec_idxs = []
            if act_couple[0] in trace:
                spec_idxs = spec_idxs + [(act_couple[0], i) for i in act_idxs[act_couple[0]]]
            if act_couple[1] in trace:
                spec_idxs = spec_idxs + [(act_couple[1], i) for i in act_idxs[act_couple[1]]]
            spec_idxs = sorted(spec_idxs, key=lambda x: (x[1], x[0]))
            while len(spec_idxs) > 1:
                if spec_idxs[1][0] != act_couple[1]:
                    del spec_idxs[0]
                else:
                    break
            is_ok = True
            for i in range(len(spec_idxs)):
                if i % 2 == 0 and (spec_idxs[i][0] != act_couple[0] or i == len(spec_idxs) - 1 or spec_idxs[i + 1][0] !=
                                   act_couple[1]):
                    is_ok = False
                    break
            if not is_ok:
                trace_dict["deviations"].append([ALTPRECEDENCE, act_couple])


def __check_chain_precedence(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                             trace_dict: Dict[str, List[Any]], act_idxs: Dict[str, List[int]],
                             parameters: Optional[Dict[Any, Any]] = None):
    if CHAINPRECEDENCE in model:
        for act_couple in model[CHAINPRECEDENCE]:
            spec_idxs = []
            if act_couple[0] in trace:
                spec_idxs = spec_idxs + [(act_couple[0], i) for i in act_idxs[act_couple[0]]]
            if act_couple[1] in trace:
                spec_idxs = spec_idxs + [(act_couple[1], i) for i in act_idxs[act_couple[1]]]
            spec_idxs = sorted(spec_idxs, key=lambda x: (x[1], x[0]))
            while len(spec_idxs) > 1:
                if spec_idxs[1][0] != act_couple[1]:
                    del spec_idxs[0]
                else:
                    break
            is_ok = True
            for i in range(len(spec_idxs)):
                if i % 2 == 0 and (spec_idxs[i][0] != act_couple[0] or i == len(spec_idxs) - 1 or spec_idxs[i + 1][0] !=
                                   act_couple[1] or spec_idxs[i + 1][1] != spec_idxs[i][1] + 1):
                    is_ok = False
                    break
            if not is_ok:
                trace_dict["deviations"].append([CHAINPRECEDENCE, act_couple])


def __check_alt_succession(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                           trace_dict: Dict[str, List[Any]], act_idxs: Dict[str, List[int]],
                           parameters: Optional[Dict[Any, Any]] = None):
    if ALTSUCCESSION in model:
        for act_couple in model[ALTSUCCESSION]:
            spec_idxs = []
            if act_couple[0] in trace:
                spec_idxs = spec_idxs + [(act_couple[0], i) for i in act_idxs[act_couple[0]]]
            if act_couple[1] in trace:
                spec_idxs = spec_idxs + [(act_couple[1], i) for i in act_idxs[act_couple[1]]]
            spec_idxs = sorted(spec_idxs, key=lambda x: (x[1], x[0]))
            is_ok = True
            for i in range(len(spec_idxs)):
                if i % 2 == 0 and (spec_idxs[i][0] != act_couple[0] or i == len(spec_idxs) - 1 or spec_idxs[i + 1][0] !=
                                   act_couple[1]):
                    is_ok = False
                    break
            if not is_ok:
                trace_dict["deviations"].append([ALTSUCCESSION, act_couple])


def __check_chain_succession(trace: List[str], model: Dict[str, Dict[Any, Dict[str, int]]],
                             trace_dict: Dict[str, List[Any]], act_idxs: Dict[str, List[int]],
                             parameters: Optional[Dict[Any, Any]] = None):
    if CHAINSUCCESSION in model:
        for act_couple in model[CHAINSUCCESSION]:
            spec_idxs = []
            if act_couple[0] in trace:
                spec_idxs = spec_idxs + [(act_couple[0], i) for i in act_idxs[act_couple[0]]]
            if act_couple[1] in trace:
                spec_idxs = spec_idxs + [(act_couple[1], i) for i in act_idxs[act_couple[1]]]
            spec_idxs = sorted(spec_idxs, key=lambda x: (x[1], x[0]))
            is_ok = True
            for i in range(len(spec_idxs)):
                if i % 2 == 0 and (spec_idxs[i][0] != act_couple[0] or i == len(spec_idxs) - 1 or spec_idxs[i + 1][0] !=
                                   act_couple[1] or spec_idxs[i + 1][1] != spec_idxs[i][1] + 1):
                    is_ok = False
                    break
            if not is_ok:
                trace_dict["deviations"].append([CHAINSUCCESSION, act_couple])


def apply_list(projected_log: List[List[str]], model: Dict[str, Dict[Any, Dict[str, int]]],
               parameters: Optional[Dict[Any, Any]] = None) -> List[Dict[str, Any]]:
    if parameters is None:
        parameters = {}

    conf_cases = []

    total_num_constraints = 0
    for k in model:
        total_num_constraints += len(model[k])

    for trace in projected_log:
        act_idxs = {}
        for i in range(len(trace)):
            if trace[i] not in act_idxs:
                act_idxs[trace[i]] = []
            act_idxs[trace[i]].append(i)

        ret = {}
        ret["no_constr_total"] = total_num_constraints
        ret["deviations"] = []

        __check_existence(trace, model, ret, parameters)
        __check_exactly_one(trace, model, ret, parameters)
        __check_init(trace, model, ret, parameters)
        __check_responded_existence(trace, model, ret, parameters)
        __check_coexistence(trace, model, ret, parameters)
        __check_non_coexistence(trace, model, ret, parameters)
        __check_response(trace, model, ret, act_idxs, parameters)
        __check_precedence(trace, model, ret, act_idxs, parameters)
        __check_succession(trace, model, ret, act_idxs, parameters)
        __check_alt_response(trace, model, ret, act_idxs, parameters)
        __check_chain_response(trace, model, ret, act_idxs, parameters)
        __check_alt_precedence(trace, model, ret, act_idxs, parameters)
        __check_chain_precedence(trace, model, ret, act_idxs, parameters)
        __check_alt_succession(trace, model, ret, act_idxs, parameters)
        __check_chain_succession(trace, model, ret, act_idxs, parameters)
        __check_absence(trace, model, ret, parameters)
        __check_non_coexistence(trace, model, ret, parameters)

        ret["no_dev_total"] = len(ret["deviations"])
        ret["dev_fitness"] = 1 - ret["no_dev_total"] / ret["no_constr_total"]
        ret["is_fit"] = ret["no_dev_total"] == 0

        conf_cases.append(ret)

    return conf_cases


def apply(log: Union[EventLog, pd.DataFrame], model: Dict[str, Dict[Any, Dict[str, int]]],
          parameters: Optional[Dict[Any, Any]] = None) -> List[Dict[str, Any]]:
    """
    Applies conformance checking against a DECLARE model.

    Paper:
    F. M. Maggi, A. J. Mooij and W. M. P. van der Aalst, "User-guided discovery of declarative process models," 2011 IEEE Symposium on Computational Intelligence and Data Mining (CIDM), Paris, France, 2011, pp. 192-199, doi: 10.1109/CIDM.2011.5949297.

    Parameters
    --------------
    log
        Event log / Pandas dataframe
    model
        DECLARE model
    parameters
        Possible parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.CASE_ID_KEY => the attribute to be used as case identifier

    Returns
    -------------
    lst_conf_res
        List containing for every case a dictionary with different keys:
        - no_constr_total => the total number of constraints of the DECLARE model
        - deviations => a list of deviations
        - no_dev_total => the total number of deviations
        - dev_fitness => the fitness (1 - no_dev_total / no_constr_total)
        - is_fit => True if the case is perfectly fit
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)

    import pm4py

    projected_log = pm4py.project_on_event_attribute(log, activity_key, case_id_key=case_id_key)

    ret = apply_list(projected_log, model, parameters=parameters)

    return ret


def get_diagnostics_dataframe(log, conf_result, parameters=None) -> pd.DataFrame:
    """
    Gets the diagnostics dataframe from a log and the results
    of DECLARE-based conformance checking

    Parameters
    --------------
    log
        Event log
    conf_result
        Results of conformance checking

    Returns
    --------------
    diagn_dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, xes_constants.DEFAULT_TRACEID_KEY)

    import pandas as pd

    diagn_stream = []

    for index in range(len(log)):
        case_id = log[index].attributes[case_id_key]

        no_dev_total = conf_result[index]["no_dev_total"]
        no_constr_total = conf_result[index]["no_constr_total"]
        dev_fitness = conf_result[index]["dev_fitness"]

        diagn_stream.append({"case_id": case_id, "no_dev_total": no_dev_total, "no_constr_total": no_constr_total,
                             "dev_fitness": dev_fitness})

    return pandas_utils.instantiate_dataframe(diagn_stream)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from collections import Counter
from typing import Tuple, Any, Counter as TCounter


class DirectlyFollowsGraph:

    def __init__(self, graph=None, start_activities=None, end_activities=None):
        if graph is None:
            graph = {}
        if start_activities is None:
            start_activities = {}
        if end_activities is None:
            end_activities = {}
        self._graph = Counter(graph)
        self._start_activities = Counter(start_activities)
        self._end_activities = Counter(end_activities)

    @property
    def graph(self) -> TCounter[Tuple[Any, Any]]:
        return self._graph

    @property
    def start_activities(self) -> TCounter[Any]:
        return self._start_activities

    @property
    def end_activities(self) -> TCounter[Any]:
        return self._end_activities

    def __repr__(self):
        return repr(self._graph)

    def __str__(self):
        return str(self._graph)

    def __iter__(self):
        yield dict(self.graph)
        yield dict(self.start_activities)
        yield dict(self.end_activities)


DFG = DirectlyFollowsGraph
